from datetime import datetime, timedelta
import pytz

from django.conf import settings
from django.core.mail import EmailMessage
from django.utils.translation import ugettext_lazy as _
from django_cron import CronJobBase, Schedule
from ldap import INVALID_SYNTAX

from .models import TeckidsGroup, TeckidsPerson
from ticdesk_org.util import filter_disallowed_persons


class CleanGroups(CronJobBase):
    RUN_EVERY_MINS = 5
    MIN_NUM_FAILURES = 3

    schedule = Schedule(run_every_mins=RUN_EVERY_MINS)
    code = 'ticdesk_account.clean_groups'

    def do(self):
        # First retrieve CNs, then iterate over those. Prevents writing back stale objects.
        cns = list(TeckidsGroup.objects.all().values_list('cn', flat=True))

        for cn in cns:
            group = TeckidsGroup.objects.get(cn=cn)
            group.clean()

            try:
                group.save()
            except INVALID_SYNTAX:
                # Uncritical because we probably hit a no-op
                pass


class CleanPersons(CronJobBase):
    RUN_EVERY_MINS = 5
    MIN_NUM_FAILURES = 3

    schedule = Schedule(run_every_mins=RUN_EVERY_MINS)
    code = 'ticdesk_account.clean_persons'

    def do(self):
        # First retrieve uids, then iterate over those. Prevents writing back stale objects.
        uids = list(TeckidsPerson.objects.all().values_list('uid', flat=True))

        for uid in uids:
            person = TeckidsPerson.objects.get(uid=uid)

            try:
                person.save()
            except INVALID_SYNTAX:
                # Uncritical because we probably hit a no-op
                pass


class BirthdayMails(CronJobBase):
    RUN_EVERY_MINS = 1440
    MIN_NUM_FAILURES = 1

    schedule = Schedule(run_every_mins=RUN_EVERY_MINS)
    code = 'ticdesk_account.birthday_mails'

    def do(self):
        # Get persons who subscribed to birthday mails
        persons = TeckidsPerson.objects.filter(setting_birthday_mail_days__gte=1).all()

        # Get minimum year of birth
        today = datetime.today().replace(tzinfo=pytz.UTC)
        min_year = TeckidsPerson.objects.filter(date_of_birth__lte=today).order_by(
            'date_of_birth').first().date_of_birth.year

        for person in persons:
            birthday_list = ''
            last = today + timedelta(days=person.setting_birthday_mail_days)

            for delta in range(0, person.setting_birthday_mail_days + 1):
                day = today + timedelta(days=delta)
                dates = [datetime(year=year, month=day.month, day=day.day, tzinfo=pytz.UTC)
                         for year in range(min_year, today.year+1)]

                birthday_persons = filter_disallowed_persons(
                    TeckidsPerson.objects.filter(
                        date_of_birth__in=dates
                    ).order_by('-date_of_birth').all(),
                    person
                )

                if birthday_persons:
                    for birthday_person in birthday_persons:
                        birthday_list += _('%s\t%s (%s Jahre)\n') % (birthday_person.date_of_birth.strftime(settings.BIRTHDAY_FORMAT),
                                                                     birthday_person.cn, str(
                            birthday_person.age_at(day))
                        )
                    birthday_list += '\n'

            if birthday_list:
                # Generate mail
                message = EmailMessage()
                message.to = [person.mail]
                message.subject = _('Geburtstage von %s bis %s') % (
                    today.strftime('%Y-%m-%d'), last.strftime('%Y-%m-%d'))
                message.body = _(
                    'Hallo %s,\n\ndie folgenden Personen haben in nächster Zeit Geburtstag:\n\n') % person.given_name
                message.body += birthday_list
                message.send()
