from datetime import datetime
import re

from dateutil.relativedelta import relativedelta
from django import forms
from django.conf import settings
from django.core import validators
from django.utils.translation import ugettext_lazy as _
import pytz

from .models import TeckidsPerson


UID_REGEX = r'^[a-z][a-z0-9]{1,19}$'
UID_MESSAGE = _(
    'Der Benutzername darf nur aus Kleinbuchstaben und Zahlen bestehen und muss mit einem Buchstaben beginnen!')

MAIL_LOCAL_REGEX = r'^[a-z][a-z0-9._-]{1,29}$'
MAIL_LOCAL_MESSAGE = _(
    'Der lokale Teil der E-Mail-Adresse  darf nur aus Kleinbuchstaben, Zahlen, Punkt, Bindestrich und Unterstrich bestehen und muss mit einem Buchstaben beginnen!')


def is_username_free(uid):
    # Sanity check the uid again to not inject unknown stuff
    if not re.match(UID_REGEX, uid):
        # Simply skip this validator - the other one on the uid field will trigger
        return

    # Check that the username is not already taken
    if TeckidsPerson.objects.filter(uid=uid) or uid in settings.DISALLOWED_UIDS:
        raise forms.ValidationError(_('Der Benutzername ist bereits vergeben!'))


def is_date_min_years_ago(min_years):
    # Generate a validator for the desired amount of years
    def _the_validator(dt):
        dt = datetime.combine(dt, datetime.min.time()).replace(tzinfo=pytz.UTC)
        today = datetime.today().replace(tzinfo=pytz.UTC)
        delta = today - dt
        years = int(delta.days / 365.2425)
        if dt > today - relativedelta(years=years):
            years -= 1
        if years < min_years:
            raise forms.ValidationError(
                _('Du musst mindestens %d Jahre alt sein! Hast du vielleicht dein Geburtsdatum falsch angegeben?') %
                min_years)
    return _the_validator


def is_mail_taken(mail):
    return bool(TeckidsPerson.objects.filter(mail_addresses__contains=mail)
                ) or mail.split('@')[0] in settings.DISALLOWED_LOCAL_PARTS


class LoginForm(forms.Form):
    uid = forms.CharField(label=_('Benutzername'), max_length=20)
    password = forms.CharField(label=_('Passwort'), widget=forms.widgets.PasswordInput)


class ChPwForm(forms.Form):
    password_old = forms.CharField(label=_('Aktuelles Passwort'),
                                   widget=forms.widgets.PasswordInput)
    password_new = forms.CharField(label=_('Neues Passwort'), widget=forms.widgets.PasswordInput, help_text=_(
        'Das Passwort darf niemand außer dir wissen und du solltest es dir gut merken können. Bitte schreibe es nicht auf. Das Passwort muss mindestens 8 Zeichen lang sein und zwei verschiedene Sorten Zeichen enthalten (Großbuchstaben, Kleinbuchstaben, Ziffern oder Sonderzeichen).'))
    password_new_repeat = forms.CharField(
        label=_('Neues Passwort wiederholen'), widget=forms.widgets.PasswordInput)

    def clean(self):
        # Ensure new password is not the same as the old one
        password_old = self.cleaned_data.get('password_old')
        password_new = self.cleaned_data.get('password_new')
        if password_old and password_old == password_new:
            self.add_error('password_new', _(
                'Das neue Passwort muss ein anderes sein als das alte!'))

        # Ensure both new passwords are identical
        password_new_repeat = self.cleaned_data.get('password_new_repeat')
        if password_new and password_new != password_new_repeat:
            self.add_error('password_new_repeat', _('Passwörter stimmen nicht überein!'))

        super(ChPwForm, self).clean()


class RegisterFormBase(forms.Form):
    # Name and identity fields
    given_name = forms.CharField(label=_('Vorname'), max_length=100)
    sn = forms.CharField(label=_('Nachname'), max_length=100)
    date_of_birth = forms.DateField(label=_('Geburtsdatum'),
                                    widget=forms.DateInput(attrs={'type': 'date'}),
                                    validators=[is_date_min_years_ago(7)])

    # Contact fields
    mail = forms.EmailField(
        label=_('E-Mail-Adresse'),
        help_text=_('Du brauchst eine E-Mail-Adresse, um wichtige Informationen zu erhalten und damit wir dich bei Problemen erreichen können. Bitte gib unbedingt deine eigene E-Mail-Adresse an. Solltest du keine eigene E-Mail-Adresse haben, kannst du dir <a href="?new_mail=1">hier</a> direkt eine E-Mail-Adresse bei uns mitregistrieren.'))
    local = forms.CharField(
        label=_('Lokaler Teil (vor dem @-Zeichen)'),
        max_length=30,
        validators=[
            validators.RegexValidator(
                MAIL_LOCAL_REGEX,
                MAIL_LOCAL_MESSAGE)],
        help_text=_(
            MAIL_LOCAL_MESSAGE +
            ' Informationen dazu, wie du deine E-Mails abrufst, findest du unter dem Menüpunkt „Dienste“ auf unserer Website.'))
    domain = forms.ChoiceField(label=_('Domain (nach dem @-Zeichen)'),
                               choices=[(_, _) for _ in settings.MAIL_DOMAINS])
    newsletter = forms.ChoiceField(label=_('Newsletter'), choices=settings.NEWSLETTER_CHOICES)

    # User account fields
    uid = forms.CharField(label=_('Benutzername'), max_length=20,
                          validators=[validators.RegexValidator(
                              UID_REGEX, UID_MESSAGE), is_username_free],
                          help_text=UID_MESSAGE)
    password = forms.CharField(label=_('Passwort'), widget=forms.widgets.PasswordInput, help_text=_(
        'Das Passwort darf niemand außer dir wissen und du solltest es dir gut merken können. Bitte schreibe es nicht auf. Das Passwort muss mindestens 8 Zeichen lang sein und zwei verschiedene Sorten Zeichen enthalten (Großbuchstaben, Kleinbuchstaben, Ziffern oder Sonderzeichen).'))
    password_repeat = forms.CharField(label=_('Passwort wiederholen'),
                                      widget=forms.widgets.PasswordInput)

    # Additional fields
    accept_tos = forms.BooleanField(label=_('Ich habe die Nutzungsbedingungen gelesen und stimme ihnen zu.'), help_text=_(
        'Bitte lies die <a href="https://www.teckids.org/pages/nutzungsbedingungen_datenschutz.html">Nutzungsbedingungen</a> aufmerksam durch!'))
    accept_data = forms.BooleanField(label=_('Ich habe die Hinweise zur Verarbeitung meiner Daten gelesen und stimme diesen zu.'), help_text=_(
        'Ich bin mit der in den <a href="https://www.teckids.org/pages/nutzungsbedingungen_datenschutz.html">Nutzungsbedingungen</a> erklärten Verarbeitung meiner Daten einverstanden und alle angegebenen Daten sind richtig. Falls ich jünger als 16 Jahre bin, sind meine Eltern hiermit ebenfalls einverstanden und ich kann dies, auf Anfrage, nachweisen (z.B. durch Herstellung eines Kontakts zu meinen Eltern).'))

    def clean(self):
        # Ensure both passwords are identical
        password = self.cleaned_data.get('password')
        password_repeat = self.cleaned_data.get('password_repeat')
        if password and password != password_repeat:
            self.add_error('password_repeat', _('Passwörter stimmen nicht überein!'))

        super(RegisterFormBase, self).clean()


class RegisterForm(RegisterFormBase):
    local = None
    domain = None


class MailAddForm(forms.Form):
    local = forms.CharField(
        label=_('Lokaler Teil (vor dem @-Zeichen)'),
        max_length=30,
        validators=[
            validators.RegexValidator(
                MAIL_LOCAL_REGEX,
                MAIL_LOCAL_MESSAGE)],
        help_text=_(
            MAIL_LOCAL_MESSAGE +
            ' Informationen dazu, wie du deine E-Mails abrufst, findest du unter dem Menüpunkt „Dienste“ auf unserer Website.'))
    domain = forms.ChoiceField(label=_('Domain (nach dem @-Zeichen)'),
                               choices=[(_, _) for _ in settings.MAIL_DOMAINS])

    def clean(self):
        # Ensure mail address is not taken
        local = self.cleaned_data.get('local')
        domain = self.cleaned_data.get('domain')
        if is_mail_taken('%s@%s' % (local, domain)):
            self.add_error('local', _('Diese E-Mail-Adresse ist bereits belegt!'))

        super(MailAddForm, self).clean()


class RegisterFormWithNewMail(RegisterFormBase):
    mail = None

    def clean(self):
        # Ensure mail address is not taken
        local = self.cleaned_data.get('local')
        domain = self.cleaned_data.get('domain')
        if is_mail_taken('%s@%s' % (local, domain)):
            self.add_error('local', _('Diese E-Mail-Adresse ist bereits belegt!'))

        super(RegisterFormWithNewMail, self).clean()
