from datetime import datetime
from hashlib import md5
import logging
import os
import pytz
from random import shuffle
import string
from time import sleep

import django.db
from django.conf import settings
from django.contrib.staticfiles import finders
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.debug import sensitive_variables
from geopy.geocoders import OpenMapQuest
from geopy.geocoders.base import GeocoderServiceError
from ldapdb.models.fields import BooleanField, CharField, DateTimeField, FloatField, ImageField, IntegerField, ListField
import ldapdb.models
from pam import pam
from redis_collections import Dict

from .util import change_kerberos_password, ssha, subscribe_mailinglist

logger = logging.getLogger(__name__)

nominatim = OpenMapQuest(api_key=getattr(settings, 'MAPQUEST_API_KEY', None),
                         user_agent=getattr(settings, 'USER_AGENT', None))
address_cache = Dict(key='ticdesk_address_cache')


class TeckidsLdapMixin(object):
    @classmethod
    def filter_by_dn(cls, dn):
        rdn = dn.split(',')[0]
        rdn_parts = {rdn_part.split('=')[0]: rdn_part.split('=')[
            1] for rdn_part in rdn.split('+')}
        return cls.objects.filter(**rdn_parts)

    def _get_objects(self, field, cls):
        objects = cls.objects.none()

        for dn in getattr(self, field):
            if not dn.endswith(self.base_dn):
                try:
                    objects |= cls.filter_by_dn(dn)
                except ObjectDoesNotExist as exception:
                    # Do not break on broken member DNs
                    logger.error('Getting %s by DN %s failed: %s',
                                 cls.__name__, dn, str(exception))
            else:
                objects |= self.filter_by_dn(dn).members_persons

        return objects.all()

    def get_image_rel(self, field='jpeg_photo', fallback=None):
        entry_dir = os.path.join(settings.MEDIA_ROOT, 'ticdesk', 'ldap',
                                 md5(self.dn.encode("utf-8")).hexdigest())

        if not os.path.isdir(entry_dir):
            os.makedirs(entry_dir)

        filename = '%s.jpg' % os.path.join(
            entry_dir, md5(field.encode("utf-8")).hexdigest())

        if getattr(self, field):
            content = getattr(self, field)
        else:
            if fallback is None:
                fallback = finders.find('people.jpg')
            with open(fallback, 'rb') as fallback_file:
                content = fallback_file.read()

        with open(filename, 'wb') as image_file:
            image_file.write(content)

        return '%s/%s/%s/%s.jpg' % ('ticdesk',
                                    'ldap',
                                    md5(self.dn.encode("utf-8")).hexdigest(),
                                    md5(field.encode("utf-8")).hexdigest())

    def get_image_url(self, field='jpeg_photo', fallback=None):
        return '%s%s' % (settings.MEDIA_URL, self.get_image_rel(field, fallback))

    def get_image_path(self, field='jpeg_photo', fallback=None):
        return '%s/%s' % (settings.MEDIA_ROOT, self.get_image_rel(field, fallback))

    def __getattr__(self, attr):
        if attr.endswith('__url'):
            return self.get_image_url(attr[:-5])
        elif attr.endswith('__path'):
            return self.get_image_path(attr[:-6])
        elif attr.endswith('__rel'):
            return self.get_image_rel(attr[:-5])

        raise AttributeError("'%s' object has no attribute '%s'" % (
            self.__class__.__name__, attr))

    def __str__(self):
        return self.cn


class TeckidsParent(ldapdb.models.Model, TeckidsLdapMixin):
    class Meta:
        managed = False
        app_label = 'ticdesk_account'
        ordering = ('sn', 'given_name')

    base_dn = 'ou=Guardians,ou=Contacts,dc=teckids,dc=org'
    object_classes = ['inetOrgPerson', 'teckidsParent']

    # Name
    cn = CharField(db_column='cn', max_length=200,
                   primary_key=True, verbose_name=_('Name'))
    given_name = CharField(db_column='givenName',
                           max_length=100, verbose_name=_('Vorname'))
    sn = CharField(db_column='sn', max_length=100, verbose_name=_('Nachname'))

    # Contact / identity details
    mail = CharField(db_column='mail', max_length=200,
                     verbose_name=_('E-Mail'))
    home_postal_address = CharField(db_column='homePostalAddress',
                                    max_length=100, verbose_name=_('Anschrift (privat)'))
    home_phone = CharField(db_column='homePhone', max_length=20,
                           verbose_name=_('Telefon (Festnetz privat)'))
    mobile = CharField(db_column='mobile', max_length=20,
                       verbose_name=_('Handy'))

    def clean(self):
        # Set default cn if not given
        if not self.cn:
            self.cn = "%s %s" % (self.given_name, self.sn)

        super(TeckidsParent, self).clean()

    def save(self, *args, **kwargs):
        self.clean()
        super(TeckidsParent, self).save(*args, **kwargs)


class TeckidsGroup(ldapdb.models.Model, TeckidsLdapMixin):
    class Meta:
        managed = False
        app_label = 'ticdesk_account'

    base_dn = 'ou=Groups,dc=teckids,dc=org'
    object_classes = ['posixGroup', 'groupOfNames', 'teckidsGroup']

    # Group name details
    cn = CharField(db_column='cn', max_length=50, primary_key=True)
    display_name = CharField(db_column='displayName', max_length=100)
    description = CharField(db_column='description', max_length=200)
    gid_number = IntegerField(db_column='gidNumber')

    # Members of this group
    members = ListField(db_column='member')
    member_uids = ListField(db_column='memberUid')
    owners = ListField(db_column='owner')

    # FIXME unify with TeckidsProject as extra_fields
    registration_fields = ListField(
        db_column='teckidsProjectRegistrationField')

    def clean(self):
        self.sync_members()

    def sync_members(self):
        # Ensure member_uids are in sync mith members
        self.member_uids = []
        for member_dn in self.members + self.owners:
            # Rely on RDN being correctly set, for performance reasons
            rdn = member_dn.split(',')[0]
            rdn_parts = rdn.split('+')

            # Find the uid part of the RDN
            for rdn_part in rdn_parts:
                if rdn_part.startswith('uid='):
                    # Add to uids list and continue with next entry
                    uid = rdn_part.split('=')[1]
                    if uid not in self.member_uids:
                        self.member_uids.append(uid)
                    break

    def add_member(self, person):
        # Add person both to groupOfNames and posixGroup part
        if person.dn not in self.members:
            self.members.append(person.dn)
        if person.uid not in self.member_uids:
            self.member_uids.append(person.uid)

    def remove_member(self, person):
        # Remove person both from groupOfNames and posixGroup part
        if person.dn in self.members:
            self.members.remove(person.dn)
        if person.uid in self.member_uids:
            self.member_uids.remove(person.uid)

    @property
    def members_persons(self):
        return self._get_objects('members', TeckidsPerson)

    @property
    def owners_persons(self):
        return self._get_objects('owners', TeckidsPerson)

    @property
    def group_type(self):
        if self.cn.startswith('team-'):
            return ('team', _('Allegemeine Teams'))
        elif self.cn.startswith('thema-'):
            return ('topic', _('Thementeams'))
        elif self.cn.startswith('news-'):
            return ('newsletter', _('Newsletter'))
        elif self.cn.startswith('project-'):
            return ('project', _('Projekte'))
        elif self.cn.startswith('projects-'):
            return ('project', _('Projekte'))
        elif self.cn in ['members', 'fritze', 'vorstand', 'vorstand-gf', 'themenwarte', 'paeda', 'unpaeda', 'aktive']:
            return ('commitee', _('Gremien'))
        return ('privilege', _('Privilegien'))


class TeckidsPerson(ldapdb.models.Model, TeckidsLdapMixin):
    class Meta:
        managed = False
        app_label = 'ticdesk_account'
        ordering = ('sn', 'given_name')

    base_dn = 'ou=People,dc=teckids,dc=org'
    object_classes = ['inetOrgPerson', 'organizationalPerson', 'teckidsPerson',
                      'posixAccount', 'shadowAccount', 'inetLocalMailRecipient', 'krb5KDCEntry', 'krb5Principal']

    # Name
    cn = CharField(db_column='cn', max_length=200,
                   primary_key=True, verbose_name=_('Name'))
    given_name = CharField(db_column='givenName',
                           max_length=100, verbose_name=_('Vorname'))
    sn = CharField(db_column='sn', max_length=100, verbose_name=_('Nachname'))

    # Contact / identity details
    mail = CharField(db_column='mail', max_length=200,
                     verbose_name=_('E-Mail'))
    home_postal_address = CharField(db_column='homePostalAddress',
                                    max_length=100, verbose_name=_('Anschrift (privat)'), blank=True, null=True)
    home_phone = CharField(db_column='homePhone', max_length=20,
                           verbose_name=_('Telefon (Festnetz privat)'), blank=True, null=True)
    mobile = CharField(db_column='mobile', max_length=20,
                       verbose_name=_('Handy'), blank=True, null=True)
    date_of_birth = DateTimeField(
        db_column='dateOfBirth', verbose_name=_('Geburtsdatum'))
    jpeg_photo = ImageField(db_column='jpegPhoto', blank=True, null=True)

    # Location
    latitude = FloatField(db_column='teckidsLatitude', blank=True, null=True)
    longitude = FloatField(db_column='teckidsLongitude', blank=True, null=True)

    # Organisation (for children: school and class)
    o = CharField(db_column='o', max_length=50,
                  verbose_name=_('Schule'), blank=True, null=True)
    ou = CharField(db_column='ou', max_length=20,
                   verbose_name=_('Klasse'), blank=True, null=True)
    l = CharField(db_column='l', max_length=50, verbose_name=_(
        'Schulort'), blank=True, null=True)

    # Account details
    uid = CharField(db_column='uid', max_length=20,
                    primary_key=True, verbose_name=_('Benutzername'))
    uid_number = IntegerField(db_column='uidNumber')
    gid_number = IntegerField(db_column='gidNumber', default=100)
    home_directory = CharField(db_column='homeDirectory', max_length=30)
    login_shell = CharField(db_column='loginShell',
                            max_length=30, default='/bin/bash')
    user_password = CharField(db_column='userPassword', max_length=100)

    # Mail service details
    mail_addresses = ListField(db_column='mailLocalAddress')

    # Meta information (mostly used on members)
    title = CharField(db_column='title', max_length=100,
                      verbose_name=_('Titel / Rolle'), blank=True, null=True)
    description = CharField(db_column='description', max_length=5000,
                            verbose_name=_('Beschreibung'), blank=True, null=True)

    # Cash information
    #cash_account = IntegerField(db_column='teckidsCashAccount', verbose_name=_('Kontonummer'))
    iban = CharField(db_column='iBAN', max_length=50,
                     verbose_name=_('IBAN'), blank=True, null=True)
    bic = CharField(db_column='bIC', max_length=50,
                    verbose_name=_('BIC'), blank=True, null=True)

    # Member information
    employee_number = IntegerField(db_column='employeeNumber', verbose_name=_(
        'Mitgliedsnummer'), blank=True, null=True)
    member_since = DateTimeField(db_column='teckidsMemberSince',
                                 verbose_name=_('Mitglied seit'), blank=True, null=True)
    member_until = DateTimeField(db_column='teckidsMemberUntil',
                                 verbose_name=_('Mitglied bis'), blank=True, null=True)

    # Keys
    ssh_keys = ListField(db_column='sshAuthorizedKey', verbose_name=_(
        'SSH-Schlüssel'), blank=True, null=True)
    pgp_key = CharField(db_column='pgpKey', verbose_name=_(
        'PGP-Schlüssel'), blank=True, null=True)

    # Group membership
    member_of = ListField(db_column='memberOf')

    # Guardians
    guardians = ListField(db_column='guardian')

    # Settings
    setting_birthday_mail_days = IntegerField(db_column='teckidsSettingsBirthdayMailDays',
                                              default=0, verbose_name=_('Geburtstags-E-Mail-Tage')
                                              )
    members_share_info = BooleanField(
        db_column='teckidsMembersShareInfo', default=False)

    krb_princ_name = CharField(db_column="krb5PrincipalName")

    krb_princ_version = IntegerField(db_column="krb5KeyVersionNumber", default=1)

    krb_kdc_flags = IntegerField(db_column="krb5KDCFlags", default=126)

    def clean(self):
        # Set default cn if not given
        if not self.cn:
            self.cn = "%s %s" % (self.given_name, self.sn)

        if not self.krb_princ_name:
            self.krb_princ_name = f"{self.uid}@TECKIDS.ORG"

        # Set default home directory if not given
        if not self.home_directory:
            self.home_directory = "/home/%s" % self.uid

        # Choose next available uid number if not provided
        if not self.uid_number:
            self.uid_number = TeckidsPerson.objects.order_by(
                '-uid_number')[0].uid_number + 1

        # Activate SASL passthrough if principal exists
#        if TeckidsKrbPrinc.objects.filter(princ_name='%s@TECKIDS.ORG' % self.uid):
#            self.set_sasl_passthrough()

        # Update coordinates if postal address is given and changed since last save
        if self.home_postal_address and (self.uid not in address_cache or
                                         address_cache[self.uid] != self.home_postal_address):
            try:
                location = nominatim.geocode(self.home_postal_address)
            except GeocoderServiceError:
                location = None

            if location:
                self.latitude, self.longitude = location.latitude, location.longitude
                address_cache[self.uid] = self.home_postal_address

        # Remove from members group if membership ended
        if self.is_member_of('members') and self.member_until and self.member_until >= datetime.today().replace(tzinfo=pytz.UTC):
            group = TeckidsGroup.objects.get(cn='members')
            group.remove_member(self)

        # Remove teckids.org addresses if not a member
#        if not self.is_member:
#            self.mail_addresses = [_ for _ in self.mail_addresses if not _.endswith('@teckids.org')]

        super(TeckidsPerson, self).clean()

    def save(self, *args, **kwargs):
        self.clean()
        super(TeckidsPerson, self).save(*args, **kwargs)

    def age_at(self, today):
        years = today.year - self.date_of_birth.year
        if (self.date_of_birth.month > today.month
            or (self.date_of_birth.month == today.month
                and self.date_of_birth.day > today.day)):
            years -= 1
        return years

    @property
    def age(self):
        return self.age_at(datetime.today().replace(tzinfo=pytz.UTC))

    def is_member_of(self, group_cn):
        return 'cn=%s,%s' % (group_cn, TeckidsGroup.base_dn) in self.member_of

    def is_owner_of(self, group_cn):
        return self in TeckidsGroup.objects.get(cn=group_cn).owners_persons

    @property
    def is_member(self):
        now = datetime.today().replace(tzinfo=pytz.UTC)
        return self.is_member_of('members') and (
            self.member_since and self.member_since <= now) and (
            not self.member_until or self.member_until >= now)

    @property
    def is_board(self):
        return self.is_member and self.is_member_of('vorstand')

    @property
    def is_paeda(self):
        return self.is_member and self.is_member_of('paeda')

    @property
    def is_orga(self):
        return self.is_member and self.is_member_of('team-orga')

    @property
    def is_supraadmin(self):
        return self.is_member and self.is_member_of('supraadmin')

    @classmethod
    @sensitive_variables('password')
    def create_person(cls, uid, gn, sn, dob, mail, password):
        # Create new LDAP object
        person = cls.objects.create(
            uid=uid, given_name=gn, sn=sn, date_of_birth=dob, mail=mail)
        if not person:
            return None

        # Generate temporary password
        # We set the real password later by changing the temporary password to the correct one
        #  through Kerberos, as to leverage passwd's password complexity checking.
#        temp_password = list(string.ascii_letters)
#        shuffle(temp_password)
#        temp_password = ''.join(temp_password)

        # Set the password in LDAP
#        person.user_password = ssha(temp_password)
#        person.save()

        # Do PAM authentication for the new user
        # On ticdesk.teckids.org, this will trigger pam_migrate to create the Kerberos principal
#        res = False
#        for tries in range(6):
#            p_conv = pam()
#            res = p_conv.authenticate(
#                '%s%s' % (uid, settings.TICDESK_USER_INITIAL_SUFFIX), temp_password, service='ticdesk')
#            if res:
#                break
#            else:
#                sleep(5)
        res = django.db.connections.all()[1].client.connection.connection.passwd_s(person.dn, oldpw=None, newpw=password)
        person.user_password = "{SASL}%s@TECKIDS.ORG" % person.uid
        if not res:
            # Now set the real password
#            res = change_kerberos_password(uid, temp_password, password)

#            if not res[0]:
                # passwd failed - remove the new user to not confuse systems!
#                person.delete()
#                return res
#        else:
            # PAM failed - remove the new user to not confuse systems!
            person.delete()
            return (False, _('Anlegen des Benutzers fehlgeschlagen. Bitte versuche es später erneut!'))
#
        # Save again, which will activate SASL passthrough in clean()
        person.save()

        return (True, person)

    def delete(self):
        # Look for Kerberos principal with asme name and delete if existing
#        principals = TeckidsKrbPrinc.objects.filter(
#            princ_name='%s@TECKIDS.ORG' % self.uid)
#        if principals.count() > 0:
#            principals.delete()

        return super(TeckidsPerson, self).delete()

    def set_sasl_passthrough(self):
        # Activate SASL pass-through for LDAP simple binds
        # This finalises SSO for this person
        self.user_password = '{SASL}%s@TECKIDS.ORG' % self.uid

    def subscribe_newsletter(self, newsletter):
        return subscribe_mailinglist('news-%s' % newsletter, self.mail)

    @property
    def groups_member(self):
        return self._get_objects('member_of', TeckidsGroup)

    @property
    def groups_owner(self):
        # We cannot use a memberOf overlay because we use nested groups, sadly…
        # FIXME This could probably be more efficient.
        return [group for group in TeckidsGroup.objects.all() if self in group.owners_persons]

    @property
    def jpeg_photo_url(self):
        return self.get_image_url('jpeg_photo')

    @property
    def home_postal_street(self):
        try:
            return self.home_postal_address.split(', ', 1)[0]
        except BaseException:
            return ''

    @property
    def home_postal_code(self):
        try:
            return self.home_postal_address.split(', ', 1)[1].split(' ', 1)[0]
        except BaseException:
            return ''

    @property
    def home_postal_place(self):
        try:
            return self.home_postal_address.split(', ', 1)[1].split(' ', 1)[1]
        except BaseException:
            return ''

    @property
    def guardians_parents(self):
        return self._get_objects('guardians', TeckidsParent)


#class TeckidsKrbPrinc(ldapdb.models.Model):
#    class Meta:
#        managed = False
#        app_label = 'ticdesk_account'
#
#    base_dn = 'ou=KerberosPrincipals,dc=teckids,dc=org'
#    object_classes = ['account', 'krb5Principal', 'krb5KDCEntry']
#
#    # Only need the principal name as all we do is validating existence
#    princ_name = CharField(db_column='krb5PrincipalName', max_length=100)
#
