import django.db
from datetime import datetime
from django.conf import settings
from django.contrib import auth
from django.contrib.auth.decorators import login_required
from django.core.mail import EmailMessage
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.utils.translation import ugettext as _
from django.views.decorators.debug import sensitive_post_parameters
from pam import pam
import pytz

from .forms import ChPwForm, LoginForm, MailAddForm, RegisterForm, RegisterFormWithNewMail
from .models import TeckidsPerson
from .util import change_kerberos_password


@sensitive_post_parameters('password')
def login(request):
    context = {}
    login_form = LoginForm()

    if request.method == 'POST':
        login_form = LoginForm(request.POST)
        if login_form.is_valid():
            user = auth.authenticate(request,
                                     username=login_form.cleaned_data['uid'],
                                     password=login_form.cleaned_data['password'])
            if user:
                auth.login(request, user)

                # Trigger PAM once to ensure Kerberos principal creation
                pam().authenticate(
                    '%s%s' %
                    (login_form.cleaned_data['uid'],
                     settings.TICDESK_USER_INITIAL_SUFFIX),
                    login_form.cleaned_data['password'],
                    service='ticdesk')

                if request.GET.get('next'):
                    next_url = request.GET['next']
                else:
                    next_url = '/'
                return redirect(next_url)
            else:
                context['error'] = _('Benutzername oder Passwort falsch')

    context['login_form'] = login_form
    return render(request, 'ticdesk_account/login.html', context)


def logout(request):
    auth.logout(request)
    return redirect('/')


@login_required
@sensitive_post_parameters('password_old', 'password_new', 'password_new_repeat')
def change_password(request):
    context = {}
    chpw_form = ChPwForm()

    if request.method == 'POST':
        chpw_form = ChPwForm(request.POST)
        if chpw_form.is_valid():
            dn = TeckidsPerson.objects.get(uid=request.user.username).dn
            ret = django.db.connections.all()[1].client.connection.connection.passwd_s(
                dn,
                oldpw=chpw_form.cleaned_data['password_old'],
                newpw=chpw_form.cleaned_data['password_new'])

            if ret:
                context['success'] = True
            else:
                context['error'] = ret

    context['chpw_form'] = chpw_form
    return render(request, 'ticdesk_account/change_password.html', context)


@sensitive_post_parameters('password', 'password_repeat')
def register(request):
    context = {}

    # Check for new mail registration and replace form class if necessary
    new_mail = request.GET.get('new_mail', 0)
    if new_mail:
        MyRegisterForm = RegisterFormWithNewMail
    else:
        MyRegisterForm = RegisterForm

    register_form = MyRegisterForm(initial={'newsletter': 'none'})

    if request.method == 'POST':
        register_form = MyRegisterForm(request.POST)
        if register_form.is_valid():
            # Assemble mail address if new mail
            if new_mail:
                register_form.cleaned_data['mail'] = '%s@%s' % (
                    register_form.cleaned_data['local'], register_form.cleaned_data['domain'])

            ret = TeckidsPerson.create_person(
                register_form.cleaned_data['uid'],
                register_form.cleaned_data['given_name'],
                register_form.cleaned_data['sn'],
                datetime.combine(register_form.cleaned_data['date_of_birth'], datetime.min.time()).replace(tzinfo=pytz.UTC),
                register_form.cleaned_data['mail'],
                register_form.cleaned_data['password'])

            if not ret[0]:
                context['registered'] = False
                context['error'] = ret[1]
            else:
                person = ret[1]

                # Add mail local address if new mail
                if new_mail:
                    person.mail_addresses = [register_form.cleaned_data['mail']]
                    person.save()

                # Do newsletter registration, if desired
                if register_form.cleaned_data['newsletter'] != 'none':
                    person.subscribe_newsletter(register_form.cleaned_data['newsletter'])

                # E-mail user and admins
                message = EmailMessage()
                message.reply_to = ['root@teckids.org']
                message.to = [register_form.cleaned_data['mail']]
                message.cc = ['root@teckids.org']
                message.subject = _(
                    'Teckids-Account %s angelegt') % register_form.cleaned_data['uid']
                message.extra_headers = {
                    'X-OTRS-CustomerUser': register_form.cleaned_data['uid']
                }
                message.body = _("""
Hallo %s,

dein Teckids-Account mit dem Benutzernamen %s wurde angelegt.

Bitte halte dich immer an die Nutzungsbedingungen:
https://www.teckids.org/de/kleingedrucktes/nutzungsbedingungen/

Bei Fragen oder Problemen schreibe bitte eine E-Mail an
verein@teckids.org.

Viel Spaß,
dein Teckids-Team
""") % (register_form.cleaned_data['given_name'], register_form.cleaned_data['uid'])
                message.send()

                context['registered'] = True

    context['register_form'] = register_form
    return render(request, 'ticdesk_account/register.html', context)


def manage_mail(request):
    context = {}
    mail_add_form = MailAddForm()

    # Get current user person
    person = TeckidsPerson.objects.get(uid=request.user.username)

    if request.method == 'POST':
        mail_add_form = MailAddForm(request.POST)
        if mail_add_form.is_valid():
            # Assemble new mail address
            local = mail_add_form.cleaned_data['local']
            domain = mail_add_form.cleaned_data['domain']
            mail = '%s@%s' % (local, domain)

            # Add new address to person
            person.mail_addresses.append(mail)
            person.save()

            # E-mail user and admins
            message = EmailMessage()
            message.reply_to = ['root@teckids.org']
            message.to = [mail]
            if person.mail and person.mail != mail:
                message.to.append(person.mail)
            message.cc = ['root@teckids.org']
            if person.mail and person.mail != mail:
                message.cc.append(person.mail)
            message.subject = _('E-Mail-Adresse %s angelegt') % mail
            message.extra_headers = {
                'X-OTRS-CustomerUser': person.uid
            }
            message.body = _("""
Hallo %s,

deine neue E-Mail-Adresse %s wurde angelegt.

Auf unserer Website findest du immer alle Informationen, um deine
E-Mails abzurufen.

Bei Fragen oder Problemen schreibe bitte eine E-Mail an
verein@teckids.org.

Viel Spaß,
dein Teckids-Team
""") % (person.given_name, mail)
            message.send()

            context['success'] = True

    context['mail_add_form'] = mail_add_form
    return render(request, 'ticdesk_account/manage_mail.html', context)


def pgpkey(request, uid):
    person = TeckidsPerson.objects.get(uid=uid)

    res = HttpResponse(person.pgp_key, content_type='application/pgp-keys; charset=us-ascii')
    res['Content-Disposition'] = 'attachment; filename=pgp_public_key_%s.asc' % uid

    return res
