from collections import OrderedDict
import re

from django import forms
from django.apps import apps
from django.utils.translation import ugettext_lazy as _
from django_starfield import Stars
from localflavor.generic.forms import IBANFormField
from localflavor.generic.countries.sepa import IBAN_SEPA_COUNTRIES
import phonenumbers

from .models import Voucher

TeckidsGroup = apps.get_model('ticdesk_account', 'TeckidsGroup')

CHANNEL_CHOICES = [
    ('none', _('Keine Angabe')),
    ('internet', _('Aus dem Internet')),
    ('school', _('Aus der Schule')),
    ('friends', _('Von Freunden')),
    ('parents', _('Durch meine Eltern')),
    ('newsletter', _('Durch den Newsletter')),
]

SEX_CHOICES = [
    ('f', _('Mädchen (weiblich)')),
    ('m', _('Junge (männlich)')),
]

COMMENT_CHOICES = [
    ('first', _('Nur Vorname')),
    ('first_age', _('Vorname und Alter')),
    ('first_last_age', _('Vorname, Nachname und Alter')),
]

LICENCE_CHOICES = [
    ('CC-BY-4.0+',
     'Creative Commons mit Namensnennung, 4.0 oder später'),
    ('CC-BY-SA-4.0+',
     'Creative Commons mit Namensnennung und Weitergabe nur unter gleichen Bedingungen, 4.0 oder später'),
]


def is_phonenumber(number):
    try:
        phonenumbers.parse(number, 'DE')
    except BaseException:
        raise forms.ValidationError(
            _('%s ist keine gültige Telefonnummer!') % number)


def is_valid_voucher_for(project_cn):
    def _is_valid_voucher(code):
        if Voucher.objects.filter(code=code, project_cn=project_cn, used=False).count() == 0:
            raise forms.ValidationError(_('Der Gutschein-Code ist ungültig.'))
    return _is_valid_voucher


def clean_phonenumber(field):
    def _clean_phonenumber(obj):
        value = obj.cleaned_data[field]

        if value:
            pn = phonenumbers.parse(value, 'DE')
            value = phonenumbers.format_number(
                pn, phonenumbers.PhoneNumberFormat.E164)

        return value
    return _clean_phonenumber


class EventRegisterForm(forms.Form):
    cn = forms.CharField(label=_('Dein Name'), disabled=True)
    date_of_birth = forms.DateField(
        label=_('Dein Geburtsdatum'), disabled=True)
    sex = forms.ChoiceField(label=_('Dein Geschlecht'), choices=SEX_CHOICES, widget=forms.RadioSelect, help_text=_(
        'Aus verschiedenen Gründen, z.B. weil wir während der Übernachtungen aus rechtlichen Gründen Geschlechtertrennung einhalten müssen, müssen wir wissen, ob du ein Junge oder Mädchen bist. Bei manchen Namen ist das nicht immer sofort zu erkennen, daher bitten wir dich, es hier anzugeben.'))
    mail = forms.EmailField(label=_('Deine E-Mail-Adresse'), disabled=True)

    street = forms.CharField(label=_('Straße und Hausnummer'))
    postal_code = forms.CharField(label=_('Postleitzahl (PLZ)'))
    place = forms.CharField(label=_('Ort'))

    mobile = forms.CharField(
        required=False,
        label=_('Deine Handynummer'),
        help_text=_('Deine Handynummer hilft uns, dich im Notfall während der Veranstaltung zu erreichen, z.B. wenn du mit deiner Gruppe alleine auf einer Konferenz o.ä. unterwegs bist. Solltest du kein Handy haben, kannst du das Feld leer lassen.'),
        validators=[is_phonenumber])

    school = forms.CharField(label=_('Schule'), help_text=_(
        'Bitte gib den Namen deiner Schule möglichst genau so ein, wie er offiziell geschrieben wird.'))
    school_place = forms.CharField(label=_('Ort der Schule'), help_text=_(
        'Gib hier den Ort (die Stadt) ein, in der deine Schule liegt (ohne Ortsteil).'))
    school_class = forms.CharField(label=_('Klasse'), help_text=_(
        'Bitte gib die Klasse an, in die du gehst (z.B. 8a).'))

    guardian_first_name = forms.CharField(label=_('Erziehungsberechtigter (Vorname)'), help_text=_(
        'Bitte gib hier den Vornamen des Erziehungsberechtigten an, der die Anmeldung mit dir ausfüllt und im Notfall während der Veranstaltung erreichbar ist.'))
    guardian_last_name = forms.CharField(label=_('Erziehungsberechtigter (Nachname)'), help_text=_(
        'Bitte gib hier den Nachnamen des Erziehungsberechtigten an, der die Anmeldung mit dir ausfüllt und im Notfall während der Veranstaltung erreichbar ist.'))
    guardian_mail = forms.EmailField(
        label=_('E-Mail-Adresse des Erziehungsberechtigten'))
    guardian_mobile = forms.CharField(
        label=_('Handynummer des Erziehungsberechtigten'),
        help_text=_(
            'Die Handynummer benötigen wir für Notfälle, falls wir deine Eltern während der Veranstaltung dringend erreichen müssen.'),
        validators=[is_phonenumber])

    channel = forms.ChoiceField(
        label=_('Wie hast du von der Veranstaltung erfahren?'), choices=CHANNEL_CHOICES)
    comment = forms.CharField(required=False, label=_(
        'Kommentar / Bemerkungen'), widget=forms.Textarea)

    voucher_code = forms.CharField(
        required=False,
        label=_('Gutschein-Code'),
        help_text=_('Wenn du einen Gutschein für die Veranstaltung hast, gib hier den Code ein. Er wird automatisch verrechnet.'))

    donation = forms.IntegerField(required=False, min_value=0, label=_('Freiwilliger Mehrbetrag'),
                                  help_text=_('Unser Verein möchte allen Kindern und Jugendlichen die Möglichkeit bieten, an unseren Veranstaltungen teilzunehmen. Oft ist der Teilnehmerbeitrag für eine Familie jedoch nicht aufzubringen. Wir halten deshalb einen Etat bereit, aus dem wir Teilnahmen fördern können, nachdem wir die Notwendigkeit und Berechtigung sorgfältig geprüft haben. Für diesen Etat sind wir auf Spenden angewiesen. Wenn Sie einen freiwilligen Mehrbetrag für diesen Etat spenden möchten, geben Sie dies bitte hier an. Ob und wenn ja in welcher Höhe gespendet wurde, wird von uns nicht dauerhaft gespeichert und auch nicht innerhalb des Vereins, z.B. an Freizeitbetreuer, weitergegeben.')
                                  )

    accept_sepa = forms.BooleanField(
        required=False,
        label=_('Per SEPA-Lastschrift bezahlen'),
        help_text=_('Eltern: Ich ermächtige den Gläubiger Teckids e.V., Rochusstr. 2-4, 53123 Bonn mit der Gläubiger-ID DE70FZT00001497650, den Teilnehmerbeitrag von meinem Konto einmalig mit dem SEPA-Basislastschriftverfahren einzuziehen. Gleichzeitig weise ich meine Bank an, die vom Teckids e.V. von meinem Konto eingezogene SEPA-Basislastschrift einzulösen.'))

    iban = IBANFormField(
        required=False,
        label=_('IBAN (für SEPA-Lastschriftverfahren)'),
        help_text=_(
            'Wenn deine Eltern per SEPA-Lastschrift bezahlen möchten, lasse sie bitte dieses Feld ausfüllen.'),
        include_countries=IBAN_SEPA_COUNTRIES)

    accept_terms = forms.BooleanField(label=_('Einverständniserklärung Eltern'), help_text=_(
        'Eltern: Mein Kind hat das Anmeldeformular gemeinsam mit mir, aber selber, ausgefüllt und ich bin mit der Teilnahme, den Nutzungsbedinungen und den AGB einverstanden. Mir ist bewusst, dass die Anmeldung verbindlich ist und ein Rücktritt nur in Ausnahmefällen mit triftigem Grund möglich ist. Darüberhinaus erkläre ich mich einverstanden, den Teilnehmerbeitrag vorab zu begleichen und erkläre mich mit den oben genannten Erstattungsrichtlinien einverstanden.'))

    accept_data = forms.BooleanField(label=_('Einverständniserklärung Datenschutz'), help_text=_(
        'Ich bin mit der in den <a href="https://www.teckids.org/kleingedrucktes/nutzungsbedingungen/">Nutzungsbedingungen</a> erklärten Verarbeitung meiner Daten einverstanden und alle angegebenen Daten sind richtig. Falls ich jünger als 16 Jahre bin, sind meine Eltern hiermit ebenfalls einverstanden und ich kann dies, auf Anfrage, nachweisen (z.B. durch Herstellung eines Kontakts zu meinen Eltern).'))

    accept_general_terms = forms.BooleanField(label=_('Einverständniserklärung AGB'), help_text=_(
        'Ich bin mit den <a href="https://www.teckids.org/de/kleingedrucktes/agb-fur-freizeiten/">AGB</a> einverstanden und habe diese gelesen.'))

    # Cleaning methods to reformat phone numbers
    clean_mobile = clean_phonenumber('mobile')
    clean_guardian_mobile = clean_phonenumber('guardian_mobile')

    def __init__(self, project, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['voucher_code'].validators.append(
            is_valid_voucher_for(project.cn))

        EventAdditionalSurveyForm.add_fields_from_ldap(
            self, project, 'guardian_first_name')


class EventAdditionalSurveyForm(forms.Form):
    def __init__(self, project, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_fields_from_ldap(project)

    def add_fields_from_ldap(self, project, before=None):
        new_fields = OrderedDict()

        for field in project.registration_fields:
            label, help_text, *choices = field.split('|')
            var = re.sub(r'[^A-Za-z0-9]|^(?=\d)', '_', label)

            if choices:
                choices_map = [(re.sub(r'[^A-Za-z0-9]|^(?=\d)', '_', choice), choice)
                               for choice in choices]
                field_attr = forms.ChoiceField(
                    label=label, help_text=help_text, choices=choices_map, required=False)
            else:
                field_attr = forms.CharField(
                    label=label, help_text=help_text, required=False)

            new_fields[var] = field_attr

        if before:
            before_field_index = list(self.fields.keys()).index(before)
            field_names = list(self.fields.keys())
            new_field_order = field_names[:before_field_index - 1] + \
                list(new_fields.keys()) + field_names[before_field_index - 1:]

        self.fields.update(new_fields)

        if before:
            self.order_fields(new_field_order)


class EventFeedbackForm(forms.Form):
    comment_private = forms.CharField(
        required=False,
        label=_('Kommentar für das Team'),
        help_text=_(
            'Dieser Kommentar ist nur für das Team. Du kannst hier alles aufschreiben, was du uns noch als Feedback mitteilen möchtest.'),
        widget=forms.Textarea)

    comment_public = forms.CharField(
        required=False,
        label=_('Kommentar für die Website'),
        help_text=_('Dieser Kommentar ist für den Bericht auf unserer Website. Erzähle ausführlich davon, was du erlebt hast, was dir gefallen hat, was du gelernt hast und allem, was dir sonst noch einfällt.'),
        widget=forms.Textarea)

    comment_public_info = forms.ChoiceField(
        label=_('Angaben beim Kommentar?'),
        choices=COMMENT_CHOICES,
        help_text=_('Mit welchen Angaben möchtest du deinen Kommentar veröffentlichen?'))

    photos = forms.ImageField(label=_('Fotos'), widget=forms.ClearableFileInput(attrs={'multiple': True}), required=False, help_text=_(
        'Wenn du Fotos für den Bericht beitragen möchtest, kannst du diese hier hochladen. Mehrere Dateien kannst du in den meisten Dateiauswahldialogen mit STRG+Klick auswählen.'))

    photos_licence = forms.ChoiceField(label=_('Foto-Lizenz'), choices=LICENCE_CHOICES,
                                       required=False,
                                       help_text=_(
                                           'Wenn du Fotos hochlädst, wähle hier eine Lizenz aus.')
                                       )

    def __init__(self, project, *args, **kwargs):
        super(EventFeedbackForm, self).__init__(*args, **kwargs)
        self._project = project

        for aspect in project.feedback_aspects:
            var = re.sub(r'[^A-Za-z0-9]|^(?=\d)', '_', aspect)
            field = forms.IntegerField(label=_('Bewertung: %s') % aspect, widget=Stars, required=False,
                                       help_text=_(
                                           '1 = nicht gut, 5 = sehr gut')
                                       )
            self.fields[var] = field


class EventCreationForm(forms.Form):
    cn = forms.SlugField(
        max_length=24,
        label=_('CN der Veranstaltung'),
        help_text=_('Zum Beispiel froglabs-2016-tuebix.'))

    display_name = forms.CharField(
        label=_('Anzeigename der Veranstaltung'),
        help_text=_('Zum Beispiel FrogLabs 2016 TUEBIX.'))

    description = forms.CharField(
        label=_('Beschreibung der Veranstaltung'),
        help_text=_(
            'Diese Beschreibung wird auf der Registrierungsseite angezeigt.'),
        widget=forms.Textarea)

    l = forms.CharField(
        label=_('Ort der Veranstaltung'),
        help_text=_('Wo findet die Veranstaltung statt? Zum Beispiel Tübingen.'))

    date_event = forms.DateField(
        label=_('Datum der Veranstaltung'),
        help_text=_('Wann findet die Veranstaltung statt? Bei mehrtägigen Veranstaltungen immer der erste Tag der Veranstaltung.'))

    date_registration = forms.DateField(
        label=_('Registrierungsdeadline'),
        help_text=_('Bis wann darf man sich zu dieser Veranstaltung registrieren?'))

    date_retraction = forms.DateField(
        label=_('Kündigungsfrist'),
        help_text=_('Bis zu diesem Datum darf man kostenlos von der Veranstaltung abmelden.'))

    cost = forms.IntegerField(
        required=True,
        label=_('Preis der Veranstaltung'),
        help_text=_('Diesen Preis müssen Teilnehmer für diese Veranstaltung bezahlen.'))

    max_participants = forms.IntegerField(
        required=True,
        label=_('Maximale Teilnehmer'),
        help_text=_('So viele Teilnehmer dürfen diese Veranstaltung besuchen.'))

    owners = forms.MultipleChoiceField(
        required=True,
        label=_('Leiter der Veranstaltung'),
        help_text=_(
            'Hier können mehrere Mitglieder eingetragen werden. Diese Mitglieder sind dann Leiter der Veranstaltung.'),
        choices=[(person.dn, person.cn) for person in TeckidsGroup.objects.get(cn='members').members_persons])

    published = forms.BooleanField(
        required=False,
        label=_('Veranstaltung veröffentlichen'),
        help_text=_('Wenn du hier einen Haken setzt, wird die Veranstaltung in „Nächste Veranstaltungen“ angezeigt.'))
