from datetime import datetime

from django.db import models
from ldapdb.models.fields import BooleanField, CharField, DateTimeField, IntegerField, ListField
import ldapdb.models
import pytz

from ticdesk_account.models import TeckidsGroup, TeckidsLdapMixin, TeckidsPerson


class Voucher(models.Model):
    class Meta:
        app_label = 'ticdesk_events'

    code = models.CharField(max_length=8, blank=True, null=True)
    project_cn = models.CharField(max_length=50)
    person_uid = models.CharField(max_length=20, blank=True, null=True)
    discount = models.IntegerField(default=100)

    used = models.BooleanField(default=False)
    used_person_uid = models.CharField(max_length=20, blank=True, null=True)
    deleted = models.BooleanField(default=False)


class RegistrationField(models.Model):
    class Meta:
        app_label = 'ticdesk_events'
        unique_together = ('project_cn', 'person_uid', 'title')

    project_cn = models.CharField(max_length=50)
    person_uid = models.CharField(max_length=20)
    title = models.CharField(max_length=30)
    content = models.TextField()

# FIXME make inheritance from TeckidsGroup work


class TeckidsProject(ldapdb.models.Model, TeckidsLdapMixin):
    class Meta:
        managed = False
        app_label = 'ticdesk_events'
        ordering = ('date_event',)

    base_dn = 'ou=Groups,dc=teckids,dc=org'
    object_classes = ['posixGroup', 'groupOfNames', 'teckidsGroup', 'teckidsProject']

    # Group name details
    cn = CharField(db_column='cn', max_length=50, primary_key=True)
    display_name = CharField(db_column='displayName', max_length=100)
    description = CharField(db_column='description', max_length=500)
    gid_number = IntegerField(db_column='gidNumber')
    published = BooleanField(db_column='teckidsPublished', default=False)

    # Members of this group
    members = ListField(db_column='member')
    member_uids = ListField(db_column='memberUid')
    owners = ListField(db_column='owner')

    l = CharField(db_column='l', max_length=50)

    date_event = DateTimeField(db_column='teckidsProjectDate')
    date_registration = DateTimeField(db_column='teckidsProjectRegistrationDeadline')
    date_retraction = DateTimeField(db_column='teckidsProjectRetractionDeadline')

    cost = IntegerField(db_column='teckidsProjectCost')
    max_participants = IntegerField(db_column='teckidsProjectMaxParticipants')

    registration_fields = ListField(db_column='teckidsProjectRegistrationField')
    feedback_aspects = ListField(db_column='teckidsProjectFeedbackAspect')

    def clean(self):
        if not self.gid_number:
            self.gid_number = TeckidsGroup.objects.order_by('-gid_number')[0].gid_number + 1

        super(TeckidsProject, self).clean()

    def save(self, *args, **kwargs):
        self.clean()
        super(TeckidsProject, self).save(*args, **kwargs)

    def can_register(self, request=None):
        now = datetime.today().replace(tzinfo=pytz.UTC)

        if request and request.user.is_authenticated:
            if Voucher.objects.filter(
                    project_cn=self.cn, person_uid=request.user.username, used=False).count() > 0:
                return True

        if len(self.members) >= self.max_participants:
            return False

        if self.date_registration:
            return self.date_registration >= now
        return self.date_event > now

    @property
    def booked_percentage(self):
        return int(len(self.members) / self.max_participants * 100)

    def sync_members(self):
        # Ensure member_uids are in sync mith members
        self.member_uids = []
        for member_dn in self.members:
            # Rely on RDN being correctly set, for performance reasons
            rdn = member_dn.split(',')[0]
            rdn_parts = rdn.split('+')

            # Find the uid part of the RDN
            for rdn_part in rdn_parts:
                if rdn_part.startswith('uid='):
                    # Add to uids list and continue with next entry
                    uid = rdn_part.split('=')[1]
                    self.member_uids.append(uid)
                    break

    def add_member(self, person):
        # Add person both to groupOfNames and posixGroup part
        if person.dn not in self.members:
            self.members.append(person.dn)
        if person.uid not in self.member_uids:
            self.member_uids.append(person.uid)

    @property
    def members_persons(self):
        return self._get_objects('members', TeckidsPerson)

    @property
    def owners_persons(self):
        return self._get_objects('owners', TeckidsPerson)
