from datetime import datetime
import json
import re

from django.apps import apps
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.core.mail import EmailMessage
from django.utils.translation import ugettext as _
from django.shortcuts import render
from django_tables2 import RequestConfig
from ldap import INVALID_SYNTAX
import pytz

from .forms import EventFeedbackForm, EventAdditionalSurveyForm, EventRegisterForm, EventCreationForm
from .models import RegistrationField, TeckidsProject, Voucher
from .tables import EventsTable, ParticipatedEventsTable, ManageEventsTable
from .util import form_to_text_table, upload_file_to_media_url

TeckidsPerson = apps.get_model('ticdesk_account', 'TeckidsPerson')
TeckidsParent = apps.get_model('ticdesk_account', 'TeckidsParent')


@login_required
def events(request):
    context = {}

    # Get all upcoming events
    now = datetime.today().replace(tzinfo=pytz.UTC)
    events = TeckidsProject.objects.filter(date_event__gte=now, published=True)

    # Build table
    events_table = EventsTable(events)
    RequestConfig(request).configure(events_table)
    context['events_table'] = events_table

    # Get all events the person participated in.
    current_person = TeckidsPerson.objects.get(uid=request.user.username)
    participated_events = TeckidsProject.objects.filter(members=current_person.dn, date_event__lte=now)

    # Build table
    participated_events_table = ParticipatedEventsTable(participated_events)
    RequestConfig(request).configure(participated_events_table)
    context['participated_events_table'] = participated_events_table

    return render(request, 'ticdesk_events/events.html', context)


@login_required
def register_event(request, cn):
    context = {}

    # Get current person and event
    current_person = TeckidsPerson.objects.get(uid=request.user.username)
    event = TeckidsProject.objects.get(cn=cn)
    context['event'] = event

    # Prepare the form
    initial = {
        'cn': current_person.cn,
        'date_of_birth': current_person.date_of_birth,
        'mail': current_person.mail,
        'street': current_person.home_postal_street,
        'postal_code': current_person.home_postal_code,
        'place': current_person.home_postal_place,
        'mobile': current_person.mobile,
        'school': current_person.o,
        'school_place': current_person.l,
        'school_class': current_person.ou,
        'channel': 'none'
    }
    register_form = EventRegisterForm(event, initial=initial)

    # Produce error if registration is not possible
    if not event.can_register(request):
        context['error'] = _('Die Anmeldung ist nicht mehr möglich!')
        return render(request, 'ticdesk_events/register_event.html', context)

    # Check whether person is already a member of the event
    if current_person.dn in event.members:
        context['error'] = _('Du bist bereits angemeldet.')
        return render(request, 'ticdesk_events/register_event.html', context)

    if request.method == 'POST':
        register_form = EventRegisterForm(event, request.POST, initial=initial)
        if register_form.is_valid():
            # Add the current person to the event
            event.add_member(current_person)
            event.save()

            # Update person information if changed
            if ('school' in register_form.changed_data or
                'school_class' in register_form.changed_data or
                'school_place' in register_form.changed_data or
                    'mobile' in register_form.changed_data):
                current_person.o = register_form.cleaned_data['school']
                current_person.ou = register_form.cleaned_data['school_class']
                current_person.l = register_form.cleaned_data['school_place']
                current_person.mobile = register_form.cleaned_data['mobile']
                current_person.save()

            # Store postal address in database
            if ('postal_code' in register_form.changed_data or
                'place' in register_form.changed_data or
                'street' in register_form.changed_data):

                message = EmailMessage()
                message.to = ['vorstand@teckids.org', ]
                message.subject = _('Änderung der Adresse von %s') % current_person.uid
                message.body = _("""Der Benutzer %s hat seine Adresse geändert. Die neue Adresse lautet: %s, %s %s. """) % (current_person.uid, register_form.cleaned_data['street'], register_form.cleaned_data['postal_code'], register_form.cleaned_data['place'])

                # Send message
                message.send()

                current_person.home_postal_address = '%s, %s %s' % (register_form.cleaned_data['street'], register_form.cleaned_data['postal_code'], register_form.cleaned_data['place'])
                current_person.save()

            # Store additional registration fields
            for field in event.registration_fields:
                label, help_text, *choices = field.split('|')
                var = re.sub(r'[^A-Za-z0-9]|^(?=\d)', '_', label)

                field = RegistrationField.objects.get_or_create(
                    project_cn=event.cn, person_uid=current_person.uid, title=label)[0]

                field.content = register_form.cleaned_data[var]
                field.save()

            # Check for existence of given guardian
            check_guardian_note = False
            if bool(TeckidsParent.objects.filter(
                given_name=register_form.cleaned_data['guardian_first_name'],
                sn=register_form.cleaned_data['guardian_last_name']
            )):
                # Guardian with name exists - check other attributes
                guardian = TeckidsParent.objects.get(
                    given_name=register_form.cleaned_data['guardian_first_name'],
                    sn=register_form.cleaned_data['guardian_last_name']
                )
                if (guardian.mail == register_form.cleaned_data['guardian_mail'] or
                    guardian.mobile == register_form.cleaned_data['guardian_mobile']
                    ):
                    # Either mail or mobile matches - update both and link
                    guardian.mail = register_form.cleaned_data['guardian_mail']
                    guardian.mobile = register_form.cleaned_data['guardian_mobile']
                    try:
                        guardian.save()
                    except INVALID_SYNTAX:
                        # Uncritical because we hit a no-op
                        pass

                    if guardian.dn not in current_person.guardians:
                        current_person.guardians.append(guardian.dn)
                        current_person.save()
                else:
                    # Automatic guardian management failed
                    check_guardian_note = True
            else:
                # Guardian does not exist - create
                guardian = TeckidsParent.objects.create(
                    given_name=register_form.cleaned_data['guardian_first_name'],
                    sn=register_form.cleaned_data['guardian_last_name'],
                    mail=register_form.cleaned_data['guardian_mail'],
                    mobile=register_form.cleaned_data['guardian_mobile']
                )

                current_person.guardians.append(guardian.dn)
                current_person.save()

            # Check discount voucher
            if register_form.cleaned_data['voucher_code']:
                try:
                    voucher = Voucher.objects.get(code=register_form.cleaned_data['voucher_code'],
                                                  project_cn=cn, used=False
                                                  )
                except ObjectDoesNotExist as e:
                    context['error'] = _('Der angegebene Gutschein existiert nicht oder ist für diese Veranstaltung nicht gültig.')
                    return render(request, 'ticdesk_events/register_event.html', context)

                voucher.used = True
                voucher.used_person_uid = current_person.uid
                voucher.save()
            else:
                voucher = None

            # Produce e-mail to registration queue
            message = EmailMessage()
            message.reply_to = (current_person.mail, register_form.cleaned_data['guardian_mail'])
            message.to = ['anmeldung@teckids.org', ]
            message.subject = _('Anmeldung: %s von %s') % (event.display_name, current_person.cn)
            message.extra_headers = {
                'X-OTRS-DynamicField-TeckidsEvent': event.display_name,
                'X-OTRS-CustomerUser': current_person.uid
            }
            message.body = ''
            if ('street' in register_form.changed_data or
                'postal_code' in register_form.changed_data or
                        'place' in register_form.changed_data
                ):
                message.body += '***** ' + _('Geänderte Anschrift') + ' *****\n\n'
            if check_guardian_note:
                message.body += '***** ' + _('Erziehungsberechtigte prüfen') + ' *****\n\n'
            message.body += form_to_text_table(register_form, 78)
            if voucher:
                message.body += '\n\n***** ' + _('Gutschein verwendet') + ' *****\n\n'
                message.body += 'Rabatt: %d %%\n' % voucher.discount

            # Attach raw form data as attachment
            message.attach('register_form.json', json.dumps(
                register_form.cleaned_data, indent=4, default=str), 'application/json')

            # Send message
            message.send()

            # Set success
            context['success'] = True

    context['register_form'] = register_form

    return render(request, 'ticdesk_events/register_event.html', context)


@login_required
def register_event_additional(request, cn):
    context = {}

    # Get current person and event
    current_person = TeckidsPerson.objects.get(uid=request.user.username)
    event = TeckidsProject.objects.get(cn=cn)
    context['event'] = event

    register_form = EventAdditionalSurveyForm(event)

    # Check whether person is already a member of the event
    if current_person.dn not in event.members:
        context['error'] = _('Du bist nicht zu dieser Veranstaltung angemeldet.')
        return render(request, 'ticdesk_events/register_event_additional.html', context)

    if request.method == 'POST':
        register_form = EventAdditionalSurveyForm(event, request.POST)
        if register_form.is_valid():
            # Store additional registration fields
            for field in event.registration_fields:
                label, help_text, *choices = field.split('|')
                var = re.sub(r'[^A-Za-z0-9]|^(?=\d)', '_', label)

                field = RegistrationField.objects.get_or_create(
                    project_cn=event.cn, person_uid=current_person.uid, title=label)[0]

                field.content = register_form.cleaned_data[var]
                field.save()

            # Produce e-mail to registration queue
            message = EmailMessage()
            message.reply_to = (current_person.mail, )
            message.to = ['anmeldung@teckids.org', ]
            message.subject = _('Zusätzliche Angaben: %s von %s') % (
                event.display_name, current_person.cn)
            message.extra_headers = {
                'X-OTRS-DynamicField-TeckidsEvent': event.display_name,
                'X-OTRS-CustomerUser': current_person.uid
            }
            message.body = ''
            message.body += form_to_text_table(register_form, 78)

            # Attach raw form data as attachment
            message.attach('register_form.json', json.dumps(
                register_form.cleaned_data, indent=4, default=str), 'application/json')

            # Send message
            message.send()

            # Set success
            context['success'] = True

    context['register_form'] = register_form

    return render(request, 'ticdesk_events/register_event_additional.html', context)


@login_required
def feedback_event(request, cn):
    context = {}

    # Get current person and event
    current_person = TeckidsPerson.objects.get(uid=request.user.username)
    event = TeckidsProject.objects.get(cn=cn)
    context['event'] = event

    # Prepare the form
    initial = {
        'comment_public_info': 'first'
    }
    feedback_form = EventFeedbackForm(event, initial=initial)

    # Check whether person is a member of the event
    if current_person.dn not in event.members:
        context['error'] = _('Du hast an dieser Veranstaltung nicht teilgenommen.')
        return render(request, 'ticdesk_events/feedback_event.html', context)

    if request.method == 'POST':
        feedback_form = EventFeedbackForm(event, request.POST, request.FILES, initial=initial)
        if feedback_form.is_valid():
            # Handle photo uploads, if any
            photo_urls = []
            for file in request.FILES.getlist('photos'):
                url = upload_file_to_media_url(file, 'ticdesk/events/feedback/photos')
                photo_urls.append(url)

            # Produce e-mail to registration queue
            message = EmailMessage()
            message.reply_to = (current_person.mail,)
            message.to = ['anmeldung@teckids.org', ]
            message.subject = _('Feedback: %s von %s') % (event.display_name, current_person.cn)
            message.extra_headers = {
                'X-OTRS-DynamicField-TeckidsEvent': event.display_name,
                'X-OTRS-CustomerUser': current_person.uid
            }
            message.body = form_to_text_table(feedback_form, 78)
            if photo_urls:
                message.body += '\n\n' + _('Fotos:') + '\n\n'
                for url in photo_urls:
                    message.body += request.build_absolute_uri(url) + '\n'

            # Attach raw form data as attachment
            message.attach('feedback_form.json', json.dumps(
                feedback_form.cleaned_data, indent=4, default=str), 'application/json')

            # Send message
            message.send()

            # Set success
            context['success'] = True

    context['feedback_form'] = feedback_form

    return render(request, 'ticdesk_events/feedback_event.html', context)


@login_required
def create_event(request):
    context = {}

    # Get current person
    current_person = TeckidsPerson.objects.get(uid=request.user.username)

    create_event_form = EventCreationForm()

    # Check whether person is a member of team-orga or supraadmin.
    if not current_person.is_orga and not current_person.is_supraadmin:
        raise PermissionDenied

    if request.method == 'POST':
        create_event_form=EventCreationForm(request.POST)
        if create_event_form.is_valid():
            create_event=TeckidsProject.objects.create(
                cn = 'project-%s' % create_event_form.cleaned_data['cn'],
                display_name = create_event_form.cleaned_data['display_name'],
                description = create_event_form.cleaned_data['description'],
                l = create_event_form.cleaned_data['l'],
                date_event = create_event_form.cleaned_data['date_event'],
                date_registration = create_event_form.cleaned_data['date_registration'],
                date_retraction = create_event_form.cleaned_data['date_retraction'],
                cost = create_event_form.cleaned_data['cost'],
                max_participants = create_event_form.cleaned_data['max_participants'],
                owners = create_event_form.cleaned_data['owners'],
                # FIXME come up with something more intelligent to prepoulate the group
                members = create_event_form.cleaned_data['owners'],
                published = create_event_form.cleaned_data['published'],
            )

            message = EmailMessage()
            message.reply_to = (current_person.mail,)
            message.to = ['orga@teckids.org', ]
            message.subject = _('Veranstaltungserstellung: %s') % create_event_form.cleaned_data['display_name']
            message.extra_headers = {
                'X-OTRS-CustomerUser': current_person.uid
            }
            message.body = ''
            message.body += form_to_text_table(create_event_form, 222)

            # Send message
            message.send()

            # Set success
            context['success'] = True

    context['create_event_form'] = create_event_form

    return render(request, 'ticdesk_events/create_event.html', context)

@login_required
def edit_event(request, cn):
    context = {}

    edit_event_form = EventCreationForm()
    current_person = TeckidsPerson.objects.get(uid=request.user.username)
    current_event = TeckidsProject.objects.get(cn=cn)

    # Prepare the form
    initial = {
        'cn': current_event.cn[8:],
        'display_name': current_event.display_name,
        'description': current_event.description,
        'l': current_event.l,
        'date_event': current_event.date_event,
        'date_registration': current_event.date_registration,
        'date_retraction': current_event.date_retraction,
        'cost': current_event.cost,
        'max_participants': current_event.max_participants,
        'owners': current_event.owners,
        'published': current_event.published,
    }

    edit_event_form=EventCreationForm(initial=initial)

    # Check whether person is a member of team-orga or supraadmin.
    if not current_person.is_supraadmin and not current_person.is_orga and not current_person.is_board:
        raise PermissionDenied

    if request.method == 'POST':
        edit_event_form=EventCreationForm(request.POST, initial=initial)
        if edit_event_form.is_valid():
            cn = 'project-%s' % edit_event_form.cleaned_data['cn'],
            display_name = edit_event_form.cleaned_data['display_name'],
            description = edit_event_form.cleaned_data['description'],
            l = edit_event_form.cleaned_data['l'],
            date_event = edit_event_form.cleaned_data['date_event'],
            date_registration = edit_event_form.cleaned_data['date_registration'],
            date_retraction = edit_event_form.cleaned_data['date_retraction'],
            cost = edit_event_form.cleaned_data['cost'],
            max_participants = edit_event_form.cleaned_data['max_participants'],
            owners = edit_event_form.cleaned_data['owners'],
            # FIXME come up with something more intelligent to prepoulate the group
            members = edit_event_form.cleaned_data['owners'],
            published = edit_event_form.cleaned_data['published'],

            current_event.save()

            # Set success
            context['success'] = True

    context['edit_event_form'] = edit_event_form

    return render(request, 'ticdesk_events/edit_event.html', context)

@login_required
def manage_events(request):
    context = {}

    current_person = TeckidsPerson.objects.get(uid=request.user.username)

    # Check whether person is a member of team-orga or supraadmin.
    if not current_person.is_supraadmin and not current_person.is_orga and not current_person.is_board:
        raise PermissionDenied

    # Get all upcoming events
    now = datetime.today().replace(tzinfo=pytz.UTC)
    events = TeckidsProject.objects.filter(date_event__gte=now)

    # Build table
    manage_events_table = ManageEventsTable(events)
    RequestConfig(request).configure(manage_events_table)
    context['manage_events_table'] = manage_events_table

    return render(request, 'ticdesk_events/manage_events.html', context)
