from django import forms
from django.utils.translation import ugettext_lazy as _
from django.apps import apps
from datetime import datetime

import pytz
import phonenumbers

TeckidsProject = apps.get_model('ticdesk_events', 'TeckidsProject')


def is_phonenumber(number):
    try:
        phonenumbers.parse(number, 'DE')
    except BaseException:
        raise forms.ValidationError(
            _('%s ist keine gültige Telefonnummer!') % number)


class AddGuardianForm(forms.Form):
    given_name = forms.CharField(label=_('Vorname'), max_length=100)
    sn = forms.CharField(label=_('Nachname'), max_length=100)
    mail = forms.EmailField(label=_('E-Mail-Adresse'))
    mobile = forms.CharField(label=_('Handynummer'), validators=[
                             is_phonenumber], required=False)
    street = forms.CharField(label=_('Straße und Hausnummer'), required=False)
    postal_code = forms.CharField(
        label=_('Postleitzahl (PLZ)'), required=False)
    place = forms.CharField(label=_('Ort'), required=False)


RECIPIENT_CHOICES = [
    ('guardians', _('Erziehungsberechtigte')),
    ('group', _('Gruppenmitglieder')),
    ('owners', _('Veranstaltungsleiter')),
]

SERVICE_CHOICES = [
    ('mail', _('E-Mail')),
    ('sms', _('SMS')),
]

SENDER_CHOICES = [
    ('self', _('Persönlich')),
    ('self_masq', _('Persönlich (nur Name)')),
    ('org', _('Verein')),
]


class GroupContactForm(forms.Form):
    service = forms.ChoiceField(label=_('Versandart'), choices=SERVICE_CHOICES)
    sender = forms.ChoiceField(label=_('Absender'), choices=SENDER_CHOICES)
    recipients = forms.MultipleChoiceField(
        label=_('Empfänger'), choices=RECIPIENT_CHOICES)
    subject = forms.CharField(label=_('Betreff'))
    text = forms.CharField(label=_('Text'), widget=forms.Textarea)


class VoucherCreationForm(forms.Form):
    project_cn = forms.ChoiceField(
        label=_('Name der Veranstaltung'),
        help_text=_(
            'Wähle die Veranstaltung aus, für die du einen Voucher erstellen möchtest.'),
        choices=[(group.cn, group.display_name) for group in TeckidsProject.objects.filter(date_event__gte=datetime.today().replace(tzinfo=pytz.UTC))])

    person_uid = forms.CharField(
        required=False,
        label=_('Benutzername der Person'),
        help_text=_('Achtung: Benutzername und nicht der richtige Name!'))

    discount = forms.IntegerField(
        label=_('Rabatt'),
        help_text=_('Rabattwert welchen die Person erhalten soll. Maximal 100.'))


class EditUserForm(forms.Form):
    # Directly editable for everyone
    mail = forms.CharField(
        required=False,
        label=_('E-Mail-Adresse'),
        help_text=_('Diese E-Mail-Adresse ist die Hauptmailadresse der Person'))

    home_phone = forms.CharField(
        required=False,
        label=_('Festnetztelefon'),
        help_text=_('Nummer des Festnetztelefons'))

    mobile = forms.CharField(
        required=False,
        label=_('Mobilfunknummer'),
        help_text=_('Nummer des Mobilfunktelefons'))

    o = forms.CharField(
        required=False,
        label=_('Schule'),
        help_text=_('Offizieller Name der Schule'))

    ou = forms.CharField(
        required=False,
        label=_('Klasse'),
        help_text=_('Klasse des Kindes. Zum Beispiel 8a'))

    l = forms.CharField(
        required=False,
        label=_('Schulort'),
        help_text=_('Ort wo die Schule steht'))

    # Directly editable only by members
    members_share_info = forms.BooleanField(
        required=False,
        label=_('Personenbezogene Daten für Mitglieder sichtbar machen.'),
        help_text=_('Personenbezogene Daten für Mitglieder sichtbar machen.'))

    iban = forms.CharField(
        required=False,
        label=_('IBAN'),
        help_text=_('IBAN deines Kontos. Zum Beispiel: DE19 1234 1234 1234 1234 12'))

    bic = forms.CharField(
        required=False,
        label=_('BIC'),
        help_text=_('BIC deines Kontos. Zum Beispiel: WELADED1WMK'))

    title = forms.CharField(
        required=False,
        label=_('Titel / Rolle'),
        help_text=_('Rolle der Person im Verein. Zum Beispiel Teamleiter PR, Systemadministrator'))

    description = forms.CharField(
        widget=forms.Textarea,
        required=False,
        label=_('Beschreibung'),
        help_text=_('z.B. Vorstellungstext'))

    setting_birthday_mail_days = forms.IntegerField(
        required=False,
        label=_('Geburtstags-E-Mails'),
        help_text=_('Gib hier ein, wie viele Tage im Voraus du wissen möchtest, wann wer Geburtstag hat. Zum Beispiel 5.'))

    # Editable by users with notification
    street = forms.CharField(label=_('Straße und Hausnummer'))
    postal_code = forms.CharField(label=_('Postleitzahl (PLZ)'))
    place = forms.CharField(label=_('Ort'))

    # Only editable by board
    jpeg_photo = forms.ImageField(label=_('Foto'), required=False)


class EditVoucherForm(forms.Form):
    code = forms.CharField(
        required=False,
        label=_('Gutscheincode'),
        disabled=True,
        help_text=_('Diesen Code muss der Teilnehmer bei der Anmeldung angeben.'))

    project_cn = forms.CharField(
        required=False,
        label=_('CN des Projekts'),
        help_text=_('Für dieses Projekt ist der Gutschein gültig'))

    person_uid = forms.CharField(
        required=False,
        label=_('Benutzername'),
        help_text=_('Für diese Person ist der Gutscheincode gültig'))

    discount = forms.IntegerField(
        required=False,
        label=_('Rabatt'),
        help_text=_('Rabattwert den die Person bekommen soll. Zum Beispiel 20.'))

    used_person_uid = forms.CharField(
        required=False,
        label=_('Genutzt von'),
        help_text=_('Von dieser Person wurde der Voucher genutzt.'))

    used = forms.BooleanField(
        required=False,
        label=_('Genutzt'),
        help_text=_('Wenn Du hier ein Häkchen setzt, wird der Voucher als benutzt markiert..'))

    deleted = forms.BooleanField(
        required=False,
        label=_('Gelöscht'),
        help_text=_('Wenn Du hier ein Häkchen setzt, wird der Voucher gelöscht.'))
