from collections import OrderedDict
import re

from django.apps import apps
from django.utils.translation import ugettext_lazy as _
from django_filters import FilterSet
import django_tables2 as tables
from django_tables2.utils import A

TeckidsPerson = apps.get_model('ticdesk_account', 'TeckidsPerson')
RegistrationField = apps.get_model('ticdesk_events', 'RegistrationField')


class PersonFilter(FilterSet):
    class Meta:
        model = TeckidsPerson
        fields = {
                  'given_name': ['contains'],
                  'sn': ['contains'],
                  'mail': ['contains'],
                  'uid': ['contains']
                 }

class PersonTable(tables.Table):
    class Meta:
        attrs = {'class': 'table table-striped table-bordered table-hover table-responsive'}

    given_name = tables.Column(verbose_name=_('Vorname'))
    sn = tables.Column(verbose_name=_('Nachname'))
    mail = tables.columns.EmailColumn(verbose_name=_('E-Mail'))
    age = tables.Column(verbose_name=_('Alter'))

    uid = tables.LinkColumn('org:person_by_uid', args=[A('uid')], verbose_name=_('Benutzername'))

    def __init__(self, *args, **kwargs):
        if 'group' in kwargs:
            kwargs['extra_columns'] = self.__class__._get_columns_from_ldap(kwargs['group'])
            del kwargs['group']
        super().__init__(*args, **kwargs)

    @classmethod
    def _get_columns_from_ldap(cls, group):
        # FIXME Move this logic somewhere else, probably somehow make a person bindable to a group temporarily
        new_fields = []

        for field in group.registration_fields:
            label, help_text, *choices = field.split('|')
            var = re.sub(r'[^A-Za-z0-9]|^(?=\d)', '_', label)

            choices_map = {}
            if choices:
                choices_map = {re.sub(r'[^A-Za-z0-9]|^(?=\d)', '_', choice): choice
                               for choice in choices}

            class _Column(tables.Column):
                def __init__(self, group, label, choices_map, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.group = group
                    self.label = label
                    self.choices_map = choices_map

                def render(self, record):
                    registration_field = RegistrationField.objects.filter(project_cn=self.group.cn,
                                                                          person_uid=record.uid, title=self.label)
                    if registration_field.count() > 0:
                        if registration_field.first().content in self.choices_map:
                            return self.choices_map[registration_field.first().content]
                        else:
                            return registration_field.first().content
                    else:
                        return ''
            field_attr = _Column(group, label, choices_map, verbose_name=label, empty_values=[])
            new_fields.append((var, field_attr))

        return new_fields

class VouchersTable(tables.Table):
    class Meta:
        attrs = {'class': 'table table-striped table-bordered table-hover table-responsive'}

    project_cn = tables.Column(verbose_name=_('Veranstaltung'))
    discount = tables.Column(verbose_name=_('Rabattwert'))
    code = tables.Column(verbose_name=_('Rabattcode'))
    person_uid = tables.Column(verbose_name=_('Person'))
    deleted = tables.LinkColumn('org:delete_voucher', args=[A('id')], verbose_name='', text=_('Löschen'))
    edit = tables.LinkColumn('org:edit_voucher', args=[A('id')], verbose_name='', text=_('Editieren'))
