from io import BytesIO
import logging
import os
import random
import string

from django.apps import apps
from django.conf import settings
from django.template.loader import get_template
from latex import build_pdf
import requests

TeckidsPerson = apps.get_model('ticdesk_account', 'TeckidsPerson')
TeckidsGroup = apps.get_model('ticdesk_account', 'TeckidsGroup')

logger = logging.getLogger(__name__)


def may_see_group(group, user):
    # Not logged-in users cannot see anything
    if not user.is_authenticated:
        return False

    # Replace user variable it the user's TeckidsPerson
    user = TeckidsPerson.objects.get(uid=user.username)

    # Admins can see everything
    if user.is_supraadmin:
        return True

    # Board members can see everything
    if user.is_board:
        return True

    # Members of pedagogic board can see everything
    if user.is_paeda:
        return True

    # Owners of the group may see the group
    if user in group.owners_persons:
        return True

    # Members of the group can see the group if also member of members
    if user.is_member and group in user.groups_member:
        return True

    # Members of the organisation can see all teams
    if user.is_member and group.group_type[0] in ('team', 'topic', 'commitee'):
        return True

    # Secure fallback to False
    logger.warning("%s tried to access disallowed group %s", user.uid, group.dn)
    return False


def may_see_person(person, user):
    if not isinstance(user, TeckidsPerson):
        # Not logged-in users cannot see anything
        if not user.is_authenticated:
            return False

        # Replace user variable with the user's TeckidsPerson
        user = TeckidsPerson.objects.get(uid=user.username)

    # Everyone can see themselves
    if user == person:
        return True

    # Admins can see everything
    if user.is_supraadmin:
        return True

    # Board members can see everything
    if user.is_board:
        return True

    # Members of pedagogic board can see everything
    if user.is_paeda:
        return True

    # Owners of groups can see their members
    if set(user.groups_owner) & set(person.groups_member):
        return True

    # Users can see each other if member of group members
    if user.is_member and person.is_member and person.members_share_info:
        return True

    # Secure fallback to False
    logger.warning("%s tried to access disallowed person %s", user.uid, person.dn)
    return False


def may_modify_person(person, user):
    # FIXME be more restrictive here
    return may_see_person(person, user)


def filter_disallowed_persons(persons, user):
    return [person for person in persons if may_see_person(person, user)]


def get_relevant_groups(user):
    # Not logged-in users cannot see anything
    if not user.is_authenticated:
        return []

    # Replace user variable with the user's TeckidsPerson
    user = TeckidsPerson.objects.get(uid=user.username)

    relevant = []
    for group in TeckidsGroup.objects.all():
        # Skip if group is a privilege - these are special
        if group.group_type[0] == 'privilege':
            continue

        # Skip if group is a newsletter - these are special
        if group.group_type[0] == 'newsletter':
            continue

        # If person is a board member, everything is relevant
        if user.is_board:
            relevant.append(group)
            continue
            
        # If person is a supraadmin member, everything is relevant
        if user.is_supraadmin:
            relevant.append(group)
            continue

        # If person is a member of the pedagogic board, everything is relevant
        if user.is_paeda:
            relevant.append(group)
            continue

        # If the person is an owner of the group, it is relevant
        if user in group.owners_persons:
            relevant.append(group)
            continue

        # Members of Teckids may see all groups they are members of
        if user.is_member and group in user.groups_member:
            relevant.append(group)
            continue

        # Members of Teckids may see all teams
        if user.is_member and group.group_type[0] in ('team', 'topic', 'commitee'):
            relevant.append(group)
            continue

    return relevant


def render_latex(template_name, context):
    # Get the template and render it
    tpl = get_template(template_name)
    latex_str = tpl.render(context).encode('utf-8')

    return latex_str


def tex_to_pdf(latex_str):
    # Generate TEXINPUTS correctly
    texinputs = settings.TEXINPUTS + [settings.MEDIA_ROOT,
                                      os.path.abspath(os.path.dirname(__file__)), '']

    # Turn LaTeX string into file object
    with BytesIO(latex_str) as latex_file:
        # Build PDF
        pdf = build_pdf(latex_file, texinputs=texinputs)

    return pdf


def send_sms(sender, recipients, body):
    api_url = 'https://gw.cmtelecom.com/v1.0/message'
    messages = {}

    messages['authentication'] = {'producttoken': settings.CM_COM_SMS_TOKEN}

    message = {}
    message['from'] = sender
    message['to'] = [{'number': number} for number in recipients]
    message['body'] = {'type': 'AUTO', 'content': body}
    messages['msg'] = [message]

    request = {'messages': messages}
    res = requests.post(api_url, json=request)

    return res

def get_group_stats(group):
    stats = {}

    stats['members'] = len(group.members)

    if getattr(group, 'date_event', None):
        ages = [person.age_at(group.date_event) for person in group.members_persons]
    else:
        ages = [person.age for person in group.members_persons]

    stats['age_avg'] = sum(ages) / len(ages)
    stats['age_range_min'] = min(ages)
    stats['age_range_max'] = max(ages)

    return stats

def generate_code():
    alphabet = string.ascii_uppercase + string.digits
    length = 8
    code = ''.join(random.choice(alphabet) for _ in range(length))

    return code
