from io import BytesIO

from PIL import Image
from resizeimage import resizeimage

from django.apps import apps
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.core.mail import EmailMessage
from django.http import Http404, HttpResponse, JsonResponse
from django.shortcuts import redirect, render
from django.template.loader import get_template
from django.urls import reverse
from django.utils.translation import ugettext as _
from django_tables2 import RequestConfig

from .forms import AddGuardianForm, GroupContactForm, VoucherCreationForm, EditUserForm, EditVoucherForm
from .tables import PersonFilter, PersonTable, VouchersTable
from .util import filter_disallowed_persons, get_relevant_groups, get_group_stats, may_see_group, may_see_person, may_modify_person, render_latex, tex_to_pdf, send_sms, generate_code
from ticdesk_events.util import form_to_text_table

TeckidsPerson = apps.get_model('ticdesk_account', 'TeckidsPerson')
TeckidsGroup = apps.get_model('ticdesk_account', 'TeckidsGroup')
TeckidsParent = apps.get_model('ticdesk_account', 'TeckidsParent')
Voucher = apps.get_model('ticdesk_events', 'Voucher')


@login_required
def person(request, uid=None):
    context = {}

    # Redirect to own listing if no uid given
    if uid is None:
        target_uid = request.GET.get('uid') or request.user.username
        return redirect(reverse('org:person_by_uid', args=[target_uid]))

    # Get person and check access
    try:
        # FIXME name collision with person()
        person = TeckidsPerson.objects.get(uid=uid)
    except ObjectDoesNotExist as e:
        # Turn not-found object into a 404 error
        raise Http404 from e

    if not may_see_person(person, request.user):
        raise PermissionDenied
    else:
        # Save person once to clean it, then add to context
        try:
            person.save()
        except BaseException:
            # Not being able to save here does not matter
            pass

        context['person'] = person

    # Add information for menu
    context['relevant_groups'] = get_relevant_groups(request.user)

    return render(request, 'ticdesk_org/person.html', context)


@login_required
def persons(request, group=None):
    # Determine desired format
    output_format = request.GET.get('format', 'html')

    # For PDF format, determine template
    if output_format == 'pdf' or output_format == 'tex':
        template = request.GET.get('template', 'list_participants')

    context = {}

    if group:
        # Determine group pre-filter
        groups = TeckidsGroup.objects.filter(cn=group)
        if groups.count() == 0:
            raise Http404
        else:
            group = groups.first()

        if not may_see_group(group, request.user):
            raise PermissionDenied
        else:
            persons = group.members_persons
            context['group'] = group

        context['stats'] = get_group_stats(group)

        # Apply search filter
        filter = PersonFilter(request.GET, queryset=persons)

        # Evaluate QuerySet and filter for allowed persons
        persons = filter_disallowed_persons(list(filter.qs), request.user)

        if output_format == 'html':
            # Build table
            person_table = PersonTable(persons, group=group)
            RequestConfig(request, paginate={
                          'per_page': 25}).configure(person_table)
            context['filter'] = filter
            context['person_table'] = person_table
        elif output_format == 'geojson':
            # Build GeoJSON dictionary
            geojson = []
            for person in persons:
                if person.latitude and person.longitude:
                    # Generate popup content from template
                    template = get_template(
                        'ticdesk_org/map_popup_content.html')
                    popup_content = template.render({'person': person})

                    # Add item to pre-GeoJSON structure
                    geojson.append({"type": "Feature",
                                    "properties": {
                                        "id": person.uid_number,
                                        "popupContent": popup_content
                                    },
                                    "geometry": {
                                        "type": "Point",
                                        "coordinates": [person.longitude, person.latitude]
                                    }})
        elif output_format == 'pdf' or output_format == 'tex':
            # Simply add persons to the context
            context['persons'] = persons

    if output_format == 'html':
        # Add information for menu
        context['relevant_groups'] = get_relevant_groups(request.user)
        return render(request, 'ticdesk_org/persons.html', context)
    elif output_format == 'geojson':
        return JsonResponse(geojson, safe=False)
    elif output_format == 'pdf' or output_format == 'tex':
        latex_str = render_latex('ticdesk_org/%s.tex' % template, context)

        if output_format == 'pdf':
            pdf = tex_to_pdf(latex_str)

            # Write HTTP response
            res = HttpResponse(pdf.data, content_type='application/pdf')
            res['Content-Disposition'] = 'attachment; filename="%s_%s.pdf"' % (
                template, group.cn)
            return res
        elif output_format == 'tex':
            # Write HTTP response
            res = HttpResponse(content_type='text/x-tex')
            res['Content-Disposition'] = 'attachment; filename="%s_%s.tex"' % (
                template, group.cn)
            res.write(latex_str)
            return res


@login_required
def persons_map(request):
    context = {}

    # Get group CN
    group_cn = request.GET.get('group_cn', '')

    # Load group
    group = TeckidsGroup.objects.get(cn=group_cn)
    if not may_see_group(group, request.user):
        raise PermissionDenied

    # Pass title for heading in template
    context['title'] = group.display_name

    # Generate GeoJSON endpoint
    context['geojson'] = '%s?%s' % (
        reverse('org:persons_by_group', args=[group_cn]), 'format=geojson')

    # Add information for menu
    context['relevant_groups'] = get_relevant_groups(request.user)

    return render(request, 'ticdesk_org/map.html', context)


@login_required
def add_guardian(request, uid):
    context = {}

    # Get person and check access
    try:
        # FIXME name collision with person()
        person = TeckidsPerson.objects.get(uid=uid)
    except ObjectDoesNotExist as e:
        # Turn not-found object into a 404 error
        raise Http404 from e

    if not may_modify_person(person, request.user):
        raise PermissionDenied
    else:
        # Save person once to clean it, then add to context
        try:
            person.save()
        except BaseException:
            # Not being able to save here does not matter
            pass

        context['person'] = person

    # Prepare the form
    initial = {
        'sn': person.sn,
        'street': person.home_postal_street,
        'postal_code': person.home_postal_code,
        'place': person.home_postal_place,
    }
    add_guardian_form = AddGuardianForm(initial=initial)

    if request.method == 'POST':
        add_guardian_form = AddGuardianForm(request.POST, initial=initial)

        if add_guardian_form.is_valid():
            # Create new TeckidsParent
            parent = TeckidsParent.objects.create(
                given_name=add_guardian_form.cleaned_data['given_name'],
                sn=add_guardian_form.cleaned_data['sn']
            )
            if add_guardian_form.cleaned_data['street']:
                parent.home_postal_address = '%s, %s %s' % (add_guardian_form.cleaned_data['street'],
                                                            add_guardian_form.cleaned_data['postal_code'],
                                                            add_guardian_form.cleaned_data['place'])
            parent.mail = add_guardian_form.cleaned_data['mail']
            parent.mobile = add_guardian_form.cleaned_data['mobile']
            parent.save()

            # Link parent to person
            person.guardians.append(parent.dn)
            person.save()

            # Redirect to person detail page
            return redirect(reverse('org:person_by_uid', args=[person.uid]))

    context['add_guardian_form'] = add_guardian_form
    return render(request, 'ticdesk_org/add_guardian.html', context)


@login_required
def group_contact(request, group):
    context = {}

    current_person = TeckidsPerson.objects.get(uid=request.user.username)

    # Get group and check access
    try:
        group = TeckidsGroup.objects.get(cn=group)
    except ObjectDoesNotExist as e:
        # Turn not-found object into a 404 error
        raise Http404 from e

    if not may_see_group(group, request.user):
        raise PermissionDenied
    else:
        # Save group once to clean it, then add to context
        try:
            group.save()
        except BaseException:
            # Not being able to save here does not matter
            pass

        context['group'] = group

    group_contact_form = GroupContactForm()

    if request.method == 'POST':
        group_contact_form = GroupContactForm(request.POST)
        if group_contact_form.is_valid():
            if group_contact_form.cleaned_data['service'] == 'mail':
                message = EmailMessage()

                if group_contact_form.cleaned_data['sender'] == 'self':
                    message.from_email = '"%s" <%s>' % (
                        current_person.cn, current_person.mail)
                elif group_contact_form.cleaned_data['sender'] == 'self_masq':
                    message.from_email = '"%s" <%s>' % (
                        current_person.cn, 'verein@teckids.org')
                elif group_contact_form.cleaned_data['sender'] == 'org':
                    message.from_email = '"Teckids e.V." <verein@teckids.org>'

                message.bcc = []
                # FIXME name collision with person()
                for person in group.members_persons:
                    if 'group' in group_contact_form.cleaned_data['recipients']:
                        message.bcc.append(person.mail)
                    if 'guardians' in group_contact_form.cleaned_data['recipients']:
                        for guardian in person.guardians_parents:
                            message.bcc.append(guardian.mail)

                if 'owners' in group_contact_form.cleaned_data['recipients']:
                    for owner in group.owners_persons:
                        message.bcc.append(owner.mail)

                message.subject = group_contact_form.cleaned_data['subject']
                message.body = group_contact_form.cleaned_data['text']

                message.send()
                context['success'] = True
            elif group_contact_form.cleaned_data['service'] == 'sms':
                if group_contact_form.cleaned_data['sender'] == 'self':
                    sender = current_person.mobile
                    if not sender:
                        sender = current_person.cn[:11]
                elif group_contact_form.cleaned_data['sender'] == 'self_masq':
                    sender = current_person.cn[:11]
                elif group_contact_form.cleaned_data['sender'] == 'org':
                    sender = 'Teckids'

                to = []
                # FIXME name collision with person()
                for person in group.members_persons:
                    if 'group' in group_contact_form.cleaned_data['recipients']:
                        to.append(person.mobile)
                    if 'guardians' in group_contact_form.cleaned_data['recipients']:
                        for guardian in person.guardians_parents:
                            to.append(guardian.mobile)

                if 'owners' in group_contact_form.cleaned_data['recipients']:
                    for owner in group.owners_persons:
                        to.append(owner.mobile)

                msg = group_contact_form.cleaned_data['text']

                ret = send_sms(sender, to, msg)
                if ret.status_code == 200:
                    context['success'] = True
                else:
                    context['error'] = _(
                        'SMS konnte nicht an den Provider übergeben werden.')

    # Reset group contact form if submission was successful
    if 'success' in context:
        group_contact_form = GroupContactForm()

    context['group_contact_form'] = group_contact_form

    # Add information for menu
    context['relevant_groups'] = get_relevant_groups(request.user)

    return render(request, 'ticdesk_org/group_contact.html', context)


@login_required
def create_voucher(request):
    context = {}

    create_voucher_form = VoucherCreationForm()
    current_person = TeckidsPerson.objects.get(uid=request.user.username)

    # Check whether person is a member of team-orga or supraadmin.
    if not current_person.is_orga and not current_person.is_supraadmin and not current_person.is_board:
        raise PermissionDenied

    if request.method == 'POST':
        create_voucher_form = VoucherCreationForm(request.POST)
        if create_voucher_form.is_valid():
            create_voucher = Voucher.objects.create(
                code=generate_code(),
                project_cn='%s' % create_voucher_form.cleaned_data['project_cn'],
                discount=create_voucher_form.cleaned_data['discount'],
            )

            if create_voucher_form.cleaned_data['person_uid']:
                create_voucher.person_uid = create_voucher_form.cleaned_data['person_uid']
                create_voucher.save()

            message = EmailMessage()
            message.reply_to = ['verein@teckids.org']
            message.to = ['orga@teckids.org', ]
            message.subject = _(
                'Voucher für %s') % create_voucher_form.cleaned_data['project_cn']
            message.body = _("""Ein Voucher für die Veranstaltung %s wurde erstellt.
Person für welche der Rabatt ausgestellt wurde: %s
Rabattwert: %s
Rabattcode: %s """) % (create_voucher_form.cleaned_data['project_cn'], create_voucher_form.cleaned_data['person_uid'], create_voucher_form.cleaned_data['discount'], create_voucher.code)

            # Send message
            message.send()

            # Set success
            context['success'] = True

    context['create_voucher_form'] = create_voucher_form

    return render(request, 'ticdesk_org/create_voucher.html', context)


@login_required
def vouchers(request):
    context = {}

    # Get all unused vouchers
    vouchers = Voucher.objects.filter(used=False, deleted=False)

    # Build table
    vouchers_table = VouchersTable(vouchers)
    RequestConfig(request).configure(vouchers_table)
    context['vouchers_table'] = vouchers_table

    return render(request, 'ticdesk_org/vouchers.html', context)


@login_required
def delete_voucher(request, voucher_id):
    context = {}

    current_person = TeckidsPerson.objects.get(uid=request.user.username)
    current_voucher = Voucher.objects.get(id=voucher_id)

    # Check whether person is a member of team-orga or supraadmin.
    if not current_person.is_orga and not current_person.is_supraadmin and not current_person.is_board:
        raise PermissionDenied

    current_voucher.deleted = True

    current_voucher.save()

    context['deleted'] = True

    return render(request, 'ticdesk_org/vouchers.html', context)


@login_required
def edit_user(request, uid):
    context = {}

    edit_user_form = EditUserForm()
    current_person = TeckidsPerson.objects.get(uid=request.user.username)
    user = TeckidsPerson.objects.get(uid=uid)

    if not current_person == user and not current_person.is_supraadmin and not current_person.is_board:
        raise PermissionDenied

    # Prepare the form
    initial = {
        'mail': user.mail,
        'street': user.home_postal_street,
        'postal_code': user.home_postal_code,
        'place': user.home_postal_place,
        'home_phone': user.home_phone,
        'mobile': user.mobile,
        'o': user.o,
        'ou': user.ou,
        'l': user.l,
        'title': user.title,
        'description': user.description,
        'members_share_info': user.members_share_info,
        'iban': user.iban,
        'bic': user.bic,
        'setting_birthday_mail_days': user.setting_birthday_mail_days,
    }

    edit_user_form = EditUserForm(initial=initial)

    if not current_person.is_board and not current_person.is_supraadmin:
        edit_user_form.fields['title'].disabled = True
        edit_user_form.fields['description'].disabled = True
        edit_user_form.fields['members_share_info'].disabled = True
        edit_user_form.fields['iban'].disabled = True
        edit_user_form.fields['bic'].disabled = True

    if request.method == 'POST':
        edit_user_form = EditUserForm(
            request.POST, request.FILES or None, initial=initial)

        if not current_person.is_board and not current_person.is_supraadmin:
            edit_user_form.fields['title'].disabled = True
            edit_user_form.fields['description'].disabled = True
            edit_user_form.fields['members_share_info'].disabled = True
            edit_user_form.fields['iban'].disabled = True
            edit_user_form.fields['bic'].disabled = True
            edit_user_form.fields['jpeg_photo'].disabled = True
        if edit_user_form.is_valid():
            # Directly editable for everyone
            user.mail = edit_user_form.cleaned_data['mail']
            user.home_phone = edit_user_form.cleaned_data['home_phone']
            user.mobile = edit_user_form.cleaned_data['mobile']
            user.o = edit_user_form.cleaned_data['o']
            user.ou = edit_user_form.cleaned_data['ou']
            user.l = edit_user_form.cleaned_data['l']
            # Directly editable only by members
            user.title = edit_user_form.cleaned_data['title']
            user.description = edit_user_form.cleaned_data['description']
            user.members_share_info = edit_user_form.cleaned_data['members_share_info']
            user.iban = edit_user_form.cleaned_data['iban']
            user.bic = edit_user_form.cleaned_data['bic']
            user.setting_birthday_mail_days = edit_user_form.cleaned_data[
                'setting_birthday_mail_days']
            # Editable only for board
            if request.FILES and 'jpeg_photo' in request.FILES:
                with Image.open(request.FILES['jpeg_photo']) as image:
                    resized_image = resizeimage.resize_width(
                        image, 300, validate=False)
                    img_bytes = BytesIO()
                    resized_image.save(img_bytes, format='JPEG')
                    user.jpeg_photo = img_bytes.getvalue()
            # Editable with notification
            if ('postal_code' in edit_user_form.changed_data or
                'place' in edit_user_form.changed_data or
                    'street' in edit_user_form.changed_data):

                user.home_postal_address = '%s, %s %s' % (
                    edit_user_form.cleaned_data['street'], edit_user_form.cleaned_data['postal_code'], edit_user_form.cleaned_data['place'])

                message = EmailMessage()
                message.to = ['vorstand@teckids.org', ]
                message.subject = _('Änderung der Adresse von %s') % user.uid
                message.body = _("""Der Benutzer %s hat seine Adresse geändert. Die neue Adresse lautet: %s, %s %s. """) % (
                    user.uid, edit_user_form.cleaned_data['street'], edit_user_form.cleaned_data['postal_code'], edit_user_form.cleaned_data['place'])

                # Send message
                message.send()

            user.save()

            # Set success
            context['success'] = True

    context['edit_user_form'] = edit_user_form

    return render(request, 'ticdesk_org/edit_user.html', context)


@login_required
def edit_voucher(request, voucher_id):
    context = {}

    edit_voucher_form = EditVoucherForm()
    current_person = TeckidsPerson.objects.get(uid=request.user.username)
    current_voucher = Voucher.objects.get(id=voucher_id)

    # Prepare the form
    initial = {
        'code': current_voucher.code,
        'project_cn': current_voucher.project_cn,
        'person_uid': current_voucher.person_uid,
        'discount': current_voucher.discount,
        'used_person_uid': current_voucher.used_person_uid,
        'used': current_voucher.used,
        'deleted': current_voucher.deleted,
    }

    edit_voucher_form = EditVoucherForm(initial=initial)

    # Check whether person is a member of team-orga or supraadmin.
    if not current_person.is_supraadmin and not current_person.is_board:
        raise PermissionDenied

    if request.method == 'POST':
        edit_voucher_form = EditVoucherForm(request.POST, initial=initial)

        if edit_voucher_form.is_valid():
            current_voucher.code = edit_voucher_form.cleaned_data['code']
            current_voucher.project_cn = edit_voucher_form.cleaned_data['project_cn']
            current_voucher.person_uid = edit_voucher_form.cleaned_data['person_uid']
            current_voucher.discount = edit_voucher_form.cleaned_data['discount']
            current_voucher.used_person_uid = edit_voucher_form.cleaned_data['used_person_uid']
            current_voucher.used = (
                'True' if edit_voucher_form.cleaned_data['used'] else 'False')
            current_voucher.deleted = (
                'True' if edit_voucher_form.cleaned_data['deleted'] else 'False')

            current_voucher.save()

            # Set success
            context['success'] = True

    context['edit_voucher_form'] = edit_voucher_form

    return render(request, 'ticdesk_org/edit_voucher.html', context)
